# -*- coding: utf-8 -*-
"""
.. contents:: :local:
    
.. autofunction:: biorefineries.biodiesel.systems.create_lipid_wash_system
.. autofunction:: biorefineries.biodiesel.systems.create_lipid_pretreatment_system
.. autofunction:: biorefineries.biodiesel.systems.create_transesterification_and_biodiesel_separation_system

References
----------
.. [1] Cortés-Peña et al. BioSTEAM: A Fast and Flexible Platform for the Design,
 Simulation, and Techno-Economic Analysis of Biorefineries under Uncertainty. 
 ACS Sustainable Chem. Eng. 2020. https://doi.org/10.1021/acssuschemeng.9b07040

.. [2] Cortés-Peña et al. Economic and Environmental Sustainability of Vegetative 
 Oil Extraction Strategies at Integrated Oilcane and Oil-sorghum Biorefineries. 
 ACS Sustainable Chem. Eng. 2022

"""
import biosteam as bst
from . import units
from . import streams as s
import numpy as np
from biosteam import SystemFactory

__all__ = (
    'create_lipid_wash_system',
    'create_lipid_pretreatment_system',
    'create_transesterification_and_biodiesel_separation_system',    
)

@SystemFactory(
    ID='lipid_wash_sys',
    ins=[s.lipid, s.lipid_wash_water],
    outs=[s.washed_lipid, s.spent_wash_water]
)
def create_lipid_wash_system(ins, outs):  
    lipid, lipid_wash_water = ins
    washed_lipid, spent_wash_water = outs
    
    # Cool the oil
    H203 = bst.HXutility('H203', lipid, T=343.15, V=0, cool_only=True)
    
    recycle = bst.Stream()
    
    # Add water to wash lipid
    T208 = bst.MixTank('T208', (H203-0, lipid_wash_water, recycle))
    T208.tau = 0.10
    
    # Centrifuge out water
    C203 = bst.LiquidsSplitCentrifuge('C203',
                                        T208-0,
                                        outs=('', spent_wash_water),
                                        split=dict(Lipid=0.99,
                                                   Water=0.01))
    
    # Vacume out water
    F201 = bst.SplitFlash('F201', T=357.15, P=10000.,
                            ins=C203-0,
                            outs=('', washed_lipid),
                            split=dict(Lipid=0.0001,
                                       Water=0.999))
    
    H204 = bst.HXutility('H204', ins=F201-0, T=320, V=0)
    P204 = bst.Pump('P204', H204-0, recycle, P=101325)
    
    # Specifications within a system
    @T208.add_specification(run=True)
    def correct_lipid_wash_water():
        ins = T208.ins
        lipid, lipid_wash_water, recycle, *others = ins
        lipid_wash_water.imol['Water'] = 0.185 * sum([i.imass['Lipid'] for i in ins]) - recycle.imol['Water']

@SystemFactory(
    ID="lipid_pretreatment_sys",
    ins=[s.crude_vegetable_oil, s.acetone, s.pure_glycerine],
    outs=[s.degummed_oil, s.polar_lipids, s.wastewater]
)
def create_lipid_pretreatment_system(ins, outs):
    crude_vegetable_oil, acetone, pure_glycerine = ins
    degummed_oil, polar_lipids, wastewater = outs
    
    # Vacume out water
    F3 = bst.SplitFlash('F3', T=357.15, P=10000.,
                          ins=crude_vegetable_oil,
                          split=dict(Lipid=0.00001,
                                     Water=1.0),)
    P10 = bst.Pump('P10', F3-1, P=101325)
    H5 = bst.HXutility('H5', ins=F3-0, T=320, V=0)
    P9 = bst.Pump('P9', H5-0, wastewater, P=101325)
    T1 = bst.StorageTank('T1', acetone, tau=7*24)
    P1 = bst.Pump('P1', T1-0, P=101325.)
    T2 = bst.StorageTank('T2', pure_glycerine, tau=7*24)
    P2 = bst.Pump('P2', T2-0, P=101325.)
    N2 = bst.Stream('N2', phase='g')
    acetone_recycle = bst.Stream()
    T3 = units.BlendingTankWithSkimming('T3', 
        [P10-0, acetone_recycle], 
    )
    @T3.add_specification(run=True)
    def adjust_acetone_flow_rate():
        total_acetone_required = T3.ins[0].F_vol
        fresh_acetone = T1.ins[0]
        recycle_acetone = sum([i.outs[0].ivol['Acetone'] for i in (P5, P6)])
        acetone_required = total_acetone_required - recycle_acetone
        if acetone_required < 0.:
            fresh_acetone.ivol['Acetone'] = 0.
            P5.outs[0].ivol['Acetone'] = total_acetone_required
        else:
            fresh_acetone.ivol['Acetone'] = acetone_required
        for i in T1.path_until(M1): i._run()
        M1._run()
    
    P3 = bst.Pump('P3', T3-1, P=101325)
    F1 = bst.Flash('F1', P3-0, V=1., P=101325)
    H1 = bst.HXutility('H1', F1-0, V=0, rigorous=True)
    P5 = bst.Pump('P5', H1-0, P=101325)
    F2 = bst.Flash('F2', T3-0, ['', polar_lipids], V=1.0, P=101325)
    H2 = bst.HXutility('H2', F2-0, V=0, rigorous=True)
    P6 = bst.Pump('P6', H2-0, P=101325)
    M1 = bst.Mixer('M1', [P1-0, P5-0, P6-0], acetone_recycle)
    # bst.mark_disjunction(acetone_recycle)
    P4 = bst.Pump('P4', F1-1, P=101325)
    hx_stream = bst.Stream()
    H3 = bst.HXprocess('H3', [P4-0, hx_stream])
    glycerol_recycle = bst.Stream()
    M2 = bst.Mixer('M2', [P2-0, glycerol_recycle])
    R1 = units.GlycerolysisReactor('R1', [H3-0, M2-0, N2])
    P7 = bst.Pump('P7', R1-1, hx_stream, P=101325.)
    @R1.add_specification
    def adjust_feed_flow_rates():
        lipid = R1.ins[0]
        required_glycerol = 1.5 * (
            + lipid.imol['FFA']
            + lipid.imol['TAG']
        ) - M2.ins[1].imol['Glycerol']
        if required_glycerol < 0.:
            T2.ins[0].imol['Glycerol'] = 0.
            M2.ins[1].imol['Glycerol'] += required_glycerol
        else:
            T2.ins[0].imol['Glycerol'] = required_glycerol
        R1.ins[2].ivol['N2'] = lipid.F_vol
        T2.run_until(P7)
        
    H4 = bst.HXutility('H4', H3-1, T=333.15, V=0)
    C1 = bst.LiquidsSplitCentrifuge('C1', H4-0, ['', glycerol_recycle],
                                    split=dict(Lipid=1.))
    P8 = bst.Pump('P8', C1-0, degummed_oil)

@SystemFactory(
    ID="transesterification_and_biodiesel_separation_sys",
    ins=[s.vegetable_oil],
    outs=[s.biodiesel, s.crude_glycerol, s.wastewater]
)
def create_transesterification_and_biodiesel_separation_system(ins, outs,
        transesterification_reactions=None,
    ):
    ### Streams ###
    
    oil, = ins
    biodiesel, crude_glycerol, wastewater = outs
    
    # Fresh methanol
    methanol = bst.Stream('methanol',
                          Methanol=1,
                          price=0.547)
    
    # Catalyst
    catalyst = bst.Stream('catalyst', 
                          NaOCH3=0.25,
                          Methanol=0.75, 
                          units='kg/hr',
                          price=2.93)
    
    # Water to remove glycerol
    biodiesel_wash_water = bst.Stream('biodiesel_wash_water', Water=13.6, T=273.15+60, 
                                      price=0.000353)
    
    HCl = bst.Stream('HCl', HCl=0.21, Water=0.79,
                      price=0.35 * 0.205) # 35% HCl by mass
    
    # Acid to neutralize catalyst after second centrifuge
    HCl1 = bst.Stream('HCl1', HCl=0.21, Water=0.79)
    
    # Acid to remove soaps after first centrifuge
    HCl2 = bst.Stream('HCl2', HCl=0.21, Water=0.79)
    
    S402 = bst.FakeSplitter('S402', ins=HCl, outs=(HCl1, HCl2))
    
    # Base to neutralize acid before distillation
    NaOH = bst.Stream('NaOH', NaOH=1, price=0.41)
    
    ### Units ###
    
    ### Biodiesel Transesterification Section ###
    
    # Aparently reactors are adiabatic, meoh coming in at 40C, lipid at 60C
    
    # From USDA Biodiesel model
    x_cat = 1.05e-05 # Catalyst molar fraction in methanol feed
    
    # Mix Recycle Methanol and Fresh Methanol
    T401 = bst.StorageTank('T401')
    P401 = bst.Pump('P401')
    
    # Storage Tank for Catalyst
    T402 = bst.StorageTank('T402')
    P402 = bst.Pump('P402')
    
    # Tank for oil
    T403 = bst.StorageTank('T403')
    T403.tau = 4
    P403 = bst.Pump('P403')
    
    # Mix Methanol and Catalyst stream
    T404 = bst.MixTank('T404')
    P404 = bst.Pump('P404')
    
    # Split Methanol/Catalyst to reactors (this is done through a process specification, so use a fake splitter)
    S401 = bst.FakeSplitter('S401')
    
    # First Reactor
    R401 = units.Transesterification('R401', 
        efficiency=0.90, excess_methanol=1., T=333.15, x_catalyst=x_cat,
        transesterification=transesterification_reactions,
    )
    
    # Centrifuge to remove glycerol
    C401 = bst.LiquidsSplitCentrifuge('C401',
                                 split=dict(Lipid=0.99,  
                                            Methanol=0.40,  
                                            Glycerol=0.06, 
                                            Biodiesel=0.999, 
                                            Water=0.40, 
                                            NaOH=0,
                                            HCl=0,
                                            NaOCH3=0.40)) 
    
    P405 = bst.Pump('P405')
    
    # Second Reactor
    R402 = units.Transesterification('R402',
        efficiency=0.90, excess_methanol=1., T=333.15, x_catalyst=x_cat,
        transesterification=transesterification_reactions,
    )
    @R402.add_specification
    def adjust_feed_to_reactors():
        R402._run()
        S401.ins[0].mix_from(S401.outs)
    
    # Centrifuge to remove glycerol
    C402 = bst.LiquidsSplitCentrifuge('C402',
                             split=dict(Lipid=0.90, 
                                        Methanol=0.10, 
                                        Glycerol=0.05, 
                                        Biodiesel=0.999, 
                                        Water=0.10, 
                                        NaOH=0,  
                                        HCl=0, 
                                        NaOCH3=0.10)) 
    
    # Acids and bases per catalyst by mol
    k1 = 0.323/1.5; k2 = 1.060/1.5; k3 = 0.04505/1.5
    @T404.add_specification
    def adjust_acid_and_base():
        T404._run()
        # Adjust according to USDA biodiesel model
        f = 0.79 / 0.21
        catalyst_mol = T404.outs[0].imol['NaOCH3']
        NaOH.imol['NaOH'] = k1 * catalyst_mol
        HCl1.imol['HCl'] = mol1 = k2 * catalyst_mol
        HCl2.imol['HCl'] = mol2 = k3 * catalyst_mol
        HCl.imol['HCl'] = mol12 = mol1 + mol2
        HCl.imol['Water'] = f * mol12
        HCl1.imol['Water'] = f * mol1
        HCl2.imol['Water'] = f * mol2
    
    ### Biodiesel Purification Section ###
    
    # Water wash
    T405 = bst.MixTank('T405')
    P406 = bst.Pump('P406')
    
    # Centrifuge out water
    C403 = bst.LiquidsRatioCentrifuge('C403',
                             K_chemicals=('Methanol', 'Glycerol'),
                             Ks=np.array([0.382, 0.183]),
                             top_solvents=('Biodiesel',),
                             top_split=(0.999,),
                             bot_solvents=('Water', 'TAG', 'NaOH', 'HCl'),
                             bot_split=(0.999, 1, 1, 1))
    
    # Vacuum dry biodiesel
    # Consider doing this by split, and keeping this unit constant
    # 290 Pa, 324 K according to USDA Biodiesel Model
    F401 = bst.SplitFlash('F401',
                    order=('Water', 'Methanol', 'Biodiesel'),
                    split=(0.9999, 0.9999, 0.00001),
                    P=10000., T=331.5, Q=0)
    F401.line = 'Vacuum dryer'
    F401.material = 'Stainless steel 304'
    P407 = bst.Pump('P407', P=101325.)
    
    ### Glycerol Purification Section ###
    
    # Condense vacuumed methanol to recycle
    H401 = bst.HXutility('H401', V=0, T=295)
    P408 = bst.Pump('P408', P=101325.)
    
    # Mix recycled streams and HCl
    T406 = bst.MixTank('T406')
    P409 = bst.Pump('P409', P=101325.)
    
    # Centrifuge out waste fat
    # assume all the lipid, free_lipid and biodiesel is washed out
    C404 = bst.LiquidsSplitCentrifuge('C404', outs=('', wastewater),
                                         order=('Methanol', 'Glycerol', 'Water'),
                                         split=(0.999, 0.999, 0.99))
    
    # Add and mix NaOH
    T407 = bst.MixTank('T407')
    P410 = bst.Pump('P410')
    
    # Methanol/Water distillation column
    D401 = bst.BinaryDistillation('D401',
                      LHK=('Methanol', 'Water'), P=101325,
                      Lr=0.999, Hr=0.99999, k=2.5,
                      is_divided=True,
                      partial_condenser=False,
                      vessel_material='Stainless steel 304',
                      tray_material='Stainless steel 304')
    
    # Save column utilities
    H402 = bst.HXprocess('H402')
    
    # Glycerol/Water flash (not a distillation column)
    chemicals = H402.chemicals
    w = 0.20/chemicals.Water.MW
    g = 0.80/chemicals.Glycerol.MW
    x_water = w/(w+g)
    # ideal_thermo = D401.thermo.ideal()
    D402 = bst.BinaryDistillation('D402',
                        LHK=('Water', 'Glycerol'),
                        k=1.25,
                        P=101325,
                        y_top=0.999999,
                        x_bot=x_water,
                        tray_material='Stainless steel 304',
                        vessel_material='Stainless steel 304',
                        partial_condenser=False)
    P413 = bst.Pump('P413', P=101325.,)
    
    @D402.add_specification
    def startup_water():
        imol = D402.ins[0].imol
        glycerol = imol['Glycerol']
        imol['Water'] = water = (800. * C402.outs[0].imol['Glycerol'] + w / g * glycerol)
        z_distillate_LK = D402.y_top
        z_bottoms_LK = D402.x_bot
        z_feed_LK = water / (water + glycerol)
        if (z_bottoms_LK < z_feed_LK < z_distillate_LK):
            D402._run()
            # Remove accumulation
            D402.outs[0].imol['Water'] = 800. * C402.outs[0].imol['Glycerol']
            D402.outs[0].T = D402.outs[0].bubble_point_at_P().T
        else:
            z_feed_LK = 0.1 * z_bottoms_LK + 0.9 * z_distillate_LK
            imol['Water'] = water = glycerol / (1 - z_feed_LK)
            D402._run()
            # Remove accumulation
            D402.outs[0].imol['Water'] = water - w / g * glycerol
            D402.outs[0].T = D402.outs[0].bubble_point_at_P().T
            
    
    # Condense recycle methanol
    H403 = bst.HXutility('H403', V=0, T=315)
    P411 = bst.Pump('P411')
    
    # Condense recycle water
    H404 = bst.HXutility('H404', V=0, T=315)
    P412 = bst.Pump('P412')
    
    # Storage tank for glycerol
    T408 = bst.StorageTank('T408', outs=crude_glycerol)
    
    # Storage tank for biodiesel
    T409 = bst.StorageTank('T409', outs=biodiesel)
    F401-1-P407-0-T409
    
    ### Biodiesel system set-up ###
    
    # Biodiesel Transesterification Section
    oil-T403-P403
    (P403-0, S401-0)-R401-0-C401
    (C401-0, S401-1)-R402-0-C402-1
    
    # Specs for product https://www.afdc.energy.gov/fuels/biodiesel_specifications.html
    # minimum spec requirements:
    #  0.050 wt % water (lower to 0.045)
    #  0.2 wt % meoh (lower to 0.15)
    #  0.02 wt % glycerol (lower to 0.015)
    #  0.4 wt % free lipids (lower to 0.35)
    
    # Find Water Flow
    @T405.add_specification(prioritize=True)
    def adjust_biodiesel_wash_water():
        total_glycerol =  (C401.outs[1].imol['Glycerol'] + R402.outs[0].imol['Glycerol'])
        wash_water = (x_water / (1 - x_water) * total_glycerol
                      - HCl1.imol['Water']
                      - NaOH.imol['Water']
                      - oil.imol['Water']
                      - HCl2.imol['Water'])
        biodiesel_wash_water.imol['Water'] = wash_water if wash_water > 0 else 0.
        T405._run()
    
    D402-0-H404-0-P412
    
    # Biodiesel wash
    (C402-0, P412-0, biodiesel_wash_water, HCl1)-T405-P406-C403
    
    # Glycerol recycle and purification section
    C403-0-F401
    F401-0-H401-P408
    C401-1-P405
    (P405-0, C402-1, C403-1, P408-0, HCl2)-T406-P409-C404
    (C404-0, NaOH)-T407-P410
    P410-0-H402-0-D401-1-D402-1-P413-0-1-H402-1-T408
                 
    
    # Mass Balance for Methanol, Recycle Methanol, and Catalyst stream
    B401 = bst.MassBalance('B401',
                           description='Adjust catalyst and methanol feed to reactors',
                           variable_inlets=[catalyst, methanol],
                           constant_inlets=[D401-0],
                           constant_outlets=[1**R401, 1**R402],
                           chemical_IDs=('Methanol', 'NaOCH3'))
    
    # Find Fresh Methanol Flow
    D401-0-B401-H403-P411 # Recycle methanol
    methanol-T401-P401  # Mix fresh and recycled methanol
    catalyst-T402-P402  # Fresh catalyst
    (P411-0, P401-0, P402-0)-T404-P404-S401  # Mix Catalyst with Methanol
    
    @B401.add_specification
    def vary_inlets():
        B401.run()
        catalyst.sink.run_until(T404)
        methanol.sink.run_until(T404)
        T404.run_until(S401)
        S401.ins[0].mix_from([1**R401, 1**R402])
    
    # Initial guess
    D402.outs[0].imol['Methanol', 'Glycerol', 'Water'] = [0.00127, 3.59e-05, 0.0244]
    
    bst.mark_disjunction(P411-0)
    


# -*- coding: utf-8 -*-
"""
Created on Thu Dec 15 20:54:39 2022

@author: yrc2
"""
from biorefineries.cellulosic import chemicals

__all__ = [*chemicals.__all__, 'create_chemicals']

from biorefineries.cellulosic.chemicals import *

create_chemicals = create_cellulosic_ethanol_chemicals


# -*- coding: utf-8 -*-
"""
"""

import biosteam as bst
from thermosteam import Stream
from biorefineries.cellulosic import units
from biorefineries.cellulosic import streams as s


__all__ = (
    'create_alkaline_pretreatment_system',
)

@bst.SystemFactory(
    ID='Alkaline_pretreatment_sys',
    ins=[s.switchgrass, s.NaOH], 
    outs=[s.pretreated_biomass, s.nanofilter_retentate],
) # DOI: 10.1002/bbb.2054; Biofuels, Bioprod. Bioref. (2019)
def create_alkaline_pretreatment_system(
        ins, outs,
        solids_loading=0.091, # Liquid solids ratio of 10
        caustic_loading=1, # g NaOH / g dry feedstock
        T_pretreatment_reactor=273.15 + 121.,
        residence_time=1,
        pretreatment_reactions=None,
    ):
    
    feedstock, NaOH = ins
    pretreated_biomass, nanofilter_retentate = outs
    warm_process_water = Stream('warm_process_water',
                              T=368.15,
                              P=4.7*101325,
                              Water=1)
    pretreatment_steam = Stream('pretreatment_steam',
                    phase='g',
                    T=268+273.15,
                    P=13*101325,
                    Water=24534+3490,
                    units='kg/hr')
    
    ### Pretreatment system
    ideal = NaOH.thermo.ideal()
    T201 = bst.StorageTank('T201', NaOH, tau=7, thermo=ideal)
    P = pretreatment_steam.chemicals['H2O'].Psat(T_pretreatment_reactor + 25)
    M202 = bst.SteamMixer('M202', (feedstock, pretreatment_steam, warm_process_water), 
                          P=P, T=T_pretreatment_reactor, liquid_IDs=['H2O', 'NaOH'], solids_loading=solids_loading, 
                          thermo=ideal, solid_IDs=('Glucan', 'Xylan', 'Arabinan', 'Mannan', 'Lignin'))
    R201 = units.PretreatmentReactorSystem('R201', M202-0, tau=residence_time,
                                           T=None, thermo=ideal,
                                           reactions=pretreatment_reactions,
                                           run_vle=False)
    H1 = bst.HXutility(200, ins=R201-1, T=368, V=0)
    P201 = units.BlowdownDischargePump('P201', H1-0, thermo=ideal)
    
    PF1 = bst.PressureFilter(300, P201-0)
    solids, liquid = PF1.outs
    M1 = bst.Mixer(300, (solids, 'wash_water'))
    
    @M1.add_specification(run=True)
    def adjust_wash_water():
        solids, wash_water = M1.ins
        wash_water.imass['Water'] = solids.F_mass
        
    C1 = bst.SolidsCentrifuge(
        300, M1-0, moisture_content=0.5,
        split=PF1.split,
    )
    C1.isplit['Glucan', 'Xylan', 'Arabinan', 'Mannan', 'Lignin'] = 0.99
    bst.HydrolyzatePump('P202', C1-0, pretreated_biomass, thermo=ideal)
    M2 = bst.Mixer(300, [C1-1, PF1-1])
    NF = bst.Nanofilter(300,
        ins=M2-0, outs=['', nanofilter_retentate],
    )
    
    M202.ins.extend([NF-0, T201-0])
    M202.caustic_loading = caustic_loading
    @M202.add_specification(run=True)
    def adjust_NaOH():
        *_, NaOH_recycle, NaOH = M202.ins
        biomass = M202.ins[0]
        NaOH_requried = M202.caustic_loading * (biomass.F_mass - biomass.imass['Water', 'Sucrose', 'Glucose'].sum())
        NaOH_recycled = NaOH_recycle.imass['NaOH']
        NaOH.imass['NaOH'] = max(
            NaOH_requried - NaOH_recycled, 0
        )
    



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import thermosteam as tmo
import biosteam as bst
from biosteam import main_flowsheet as f
from biosteam import SystemFactory
from biorefineries.biodiesel import (
    create_lipid_pretreatment_system as create_oil_pretreatment_system,
    create_transesterification_and_biodiesel_separation_system,
)
from .fermentation import create_cane_to_combined_1_and_2g_fermentation
from .lipid_extraction import create_post_fermentation_oil_separation_system
from ..data import microbial_oil_baseline as perf
from .. import units
from .. import streams as s

__all__ = (
    'create_oilcane_to_biodiesel_combined_1_and_2g_post_fermentation_oil_separation',
)

@SystemFactory(
    ID='oilcane_sys',
    ins=[s.oilcane],
    outs=[s.biodiesel, s.crude_glycerol],
)
def create_oilcane_to_biodiesel_combined_1_and_2g_post_fermentation_oil_separation(
        ins, outs, fed_batch=True, WWT_kwargs=None,
        oil_extraction=None,
    ):
    if oil_extraction is None: oil_extraction = 'integrated'
    oilcane, = ins
    biodiesel, crude_glycerol = outs
    if fed_batch:
        biomass_coeff = perf.fed_batch_biomass_growth_coefficient_mean
        lipid_yield = perf.fed_batch_lipid_yield_mean
        titer = perf.fed_batch_titer_mean
        productivity = perf.fed_batch_productivity_mean
    else:
        biomass_coeff = perf.batch_biomass_growth_coefficient_mean
        lipid_yield = perf.batch_lipid_yield_mean
        titer = perf.batch_titer_mean
        productivity = perf.batch_productivity_mean
    
    glucose_fermrxn = tmo.Rxn('O2 + Glucose -> H2O + TAG', 'Glucose', 1., correct_atomic_balance=True)
    glucose_fermrxn.product_yield('TAG', basis='wt', product_yield=lipid_yield)
    xylose_fermrxn = tmo.Rxn('O2 + Xylose -> H2O + TAG', 'Xylose', 1., correct_atomic_balance=True)
    xylose_fermrxn.product_yield('TAG', basis='wt', product_yield=lipid_yield)
    cellmass_rxn = tmo.Rxn(
        'Glucose + Urea -> Yeast + H2O + CO2', 'Glucose', 1., 
        correct_atomic_balance=True
    )
    cellmass_rxn.product_yield('Yeast', 'wt', biomass_coeff)
    combustion = tmo.Rxn('Glucose + O2 -> CO2 + H2O', 'Glucose', 1. - cellmass_rxn.X,
                         correct_atomic_balance=True)
    glucose_growrxn = cellmass_rxn + combustion
    glucose_growrxn.X = 0.999 - glucose_fermrxn.X
    
    cellmass_rxn = tmo.Rxn(
        'Xylose + Urea -> Yeast + H2O + CO2', 'Xylose', 1., 
        correct_atomic_balance=True
    )
    cellmass_rxn.product_yield('Yeast', 'wt', biomass_coeff)
    combustion = tmo.Rxn('Xylose + O2 -> CO2 + H2O', 'Xylose', 1. - cellmass_rxn.X,
                         correct_atomic_balance=True)
    xylose_growrxn = cellmass_rxn + combustion
    xylose_growrxn.X = 0.999 - xylose_fermrxn.X
    cofermentation = tmo.PRxn(
        [glucose_fermrxn,
         xylose_fermrxn,
         glucose_growrxn,
         xylose_growrxn],
    )
    oilcane_to_fermentation_sys, ofs_dct = create_cane_to_combined_1_and_2g_fermentation('oilcane_to_fermentation_sys',
        ins=oilcane, 
        product_group='Lipid',
        titer=titer,
        productivity=productivity,
        cofermentation_reactions=cofermentation,
        seed_train_reactions=bst.Rxn(None, 'Glucose', 1.), # Easier to simulate reactions only at cofermentation reactor
        CoFermentation=units.AeratedCoFermentation,
        SeedTrain=units.SeedTrain,
        include_scrubber=False,
        fed_batch=fed_batch,
        udct=True,
        mockup=True
    )
    beer, lignin, condensate, pretreatment_wastewater, fiber_fines, bagasse_to_boiler = oilcane_to_fermentation_sys.outs
    hydrolysate_and_juice_mixer = bst.F.hydrolysate_and_juice_mixer
    post_fermentation_oil_separation_sys, pfls_dct = create_post_fermentation_oil_separation_system(
        ins=beer,
        mockup=True,
        area=400,
        udct=True,
        separate_cellmass=True,
    )
    backend_oil, cellmass, wastewater, evaporator_condensate = post_fermentation_oil_separation_sys.outs
    backend_oil.ID = 'backend_oil'
    # Upstream recycle of cell mass to pretreatment for oil recovery (process intensification)
    biomass_mixer = bst.Mixer(400, cellmass)
    steam_mixer = ofs_dct['Steam mixer']
    biomass_mixer.insert(steam_mixer.ins[0])
    
    # Recover all lipids in solution (after removing solids) post-fermentation
    oil_centrifuge = pfls_dct['Liquids centrifuge']
    oil_centrifuge.isplit['Oil'] = 1.0
    
    # Cells contain all the oil
    cellmass_centrifuge = pfls_dct['Solids centrifuge']
    cellmass_centrifuge.isplit['Oil'] = 0.999 # Equivalent to cell mass split
    
    
    if oil_extraction == 'integrated':
        # Move recovery of oil in solution to after hydrolysis and before fermentation (do not include juice, which may not have much oil)
        aqueous_stream = oil_centrifuge.outs[1]
        fermentation_effluent = oil_centrifuge.ins[0]
        segment = bst.Segment(fermentation_effluent, aqueous_stream)
        segment.disconnect(join_ends=True)
        segment.insert(hydrolysate_and_juice_mixer.ins[0])
        oil = backend_oil
    elif oil_extraction == 'screwpress':
        cellmass, aqueous_stream = cellmass_centrifuge.outs
        mixer = cellmass.sink
        cellmass.disconnect_sink()
        mixer.disconnect(join_ends=True)
        U402 = bst.DrumDryer(400, 
            (cellmass, 'dryer_air', 'dryer_natural_gas'), 
            ('', 'dryer_outlet_air', 'dryer_emissions'),
            moisture_content=0.18, split=0.,
            utility_agent='Steam',
        )
        # X401 = bst.ThermalOxidizer('X401', (U403-1, 'oxidizer_air'), 'oxidizer_emissions')
        U403 = bst.ScrewPress(400, U402-0, split=dict(cellmass=1, lipid=0.3, Water=0.8),)
        bst.ConveyingBelt(400, U403-0, 'yeast_extract')
        microbial_oil = U403-1
        mixer.ins[:] = [microbial_oil, backend_oil]
        oil = mixer.outs[0]
        
        # Replace oil and cell mass centrifuge with a 3-phase decanter centrifuge
        decanter = bst.SolidLiquidsSplitCentrifuge(400,
            ins=oil_centrifuge.ins[0],
            outs=[backend_oil, aqueous_stream, cellmass],
            solids_split=cellmass_centrifuge.split,
            aqueous_split=1. - oil_centrifuge.split,
            moisture_content=0.4,
        )
        oil_centrifuge.disconnect(discard=True)
        cellmass_centrifuge.disconnect(discard=True)
    else:
        raise ValueError(
            f"oil_extraction must be either 'integrated' or 'extrussion', not {repr(oil_extraction)}"
        )
    
    oil_pretreatment_sys, oil_pretreatment_dct = create_oil_pretreatment_system(
        ins=oil,
        outs=['', 'polar_lipids', ''],
        mockup=True,
        area=600,
        udct=True
    )
    pretreated_oil, polar_lipids, wastewater_small = oil_pretreatment_sys.outs
    
    create_transesterification_and_biodiesel_separation_system(
        ins=pretreated_oil, 
        outs=[biodiesel, crude_glycerol, ''],
        mockup=True,
        area=600,
    )
    
    s = f.stream
    u = f.unit
    if WWT_kwargs is None:
        WWT_kwargs = dict(area=500)
    else:
        WWT_kwargs['area'] = 500
        
    bst.create_all_facilities(
        feedstock=s.bagasse,
        recycle_process_water_streams=(condensate, evaporator_condensate),
        HXN_kwargs=dict(
            ID=900,
            ignored=lambda: [u.H401, u.D601.boiler, u.D602.boiler, u.H603, u.H602, u.H601, u.H604, u.H606, u.H609, oil_pretreatment_dct['F3']],
            Qmin=1e3,
            acceptable_energy_balance_error=0.01,
        ),
        CHP_kwargs=dict(area=700),
        WWT_kwargs=WWT_kwargs,
        area=800,
    )



# -*- coding: utf-8 -*-
"""
"""

import biosteam as bst
from thermosteam import Stream
from biorefineries.cellulosic import streams as s
from biorefineries.cellulosic import units


__all__ = (
    'create_hot_water_pretreatment_system',
)

@bst.SystemFactory(
    ID='hot_water_pretreatment_sys',
    ins=[s.cornstover],
    outs=[s.pretreated_biomass,
          s.pretreatment_wastewater],
)
def create_hot_water_pretreatment_system(
        ins, outs,
        pretreatment_area=200,
        solids_loading=0.305,
        nonsolids=['Water'],
        milling=False,
        T_pretreatment_reactor=273.15 + 130.
    ):
    
    feedstock, = ins
    pretreated_biomass, pretreatment_wastewater = outs
    
    warm_process_water = Stream('warm_process_water',
                              T=368.15,
                              P=4.7*101325,
                              Water=1)
    pretreatment_steam = Stream('pretreatment_steam',
                    phase='g',
                    T=268+273.15,
                    P=13*101325,
                    Water=24534+3490,
                    units='kg/hr')
    
    ### Pretreatment system
    n = pretreatment_area
    P = pretreatment_steam.chemicals['H2O'].Psat(T_pretreatment_reactor + 25)
    M203 = bst.SteamMixer(f'M{n+2}', (feedstock, pretreatment_steam, warm_process_water),
                          P=P, solids_loading=solids_loading)
    R201 = units.PretreatmentReactorSystem(f'R{n+1}', M203-0, T=T_pretreatment_reactor)
    P201 = units.BlowdownDischargePump(f'P{n+1}', R201-1)
    F201 = units.PretreatmentFlash(f'F{n+1}', P201-0, P=101325, Q=0)
    M204 = bst.Mixer(f'M{n+3}', (R201-0, F201-0))
    units.WasteVaporCondenser(f'H{n+1}', M204-0, pretreatment_wastewater, V=0)
    P202 = units.HydrolyzatePump(f'P{n+2}', F201-1, None)
    if milling:
        bst.HammerMill(f'U{n+1}', P202-0, pretreated_biomass)
    else:
        P202.outs[0] = pretreated_biomass



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import (
    units,
    _chemicals,
    systems,
    _tea,
    utils,
    _evaluation,
    _uncertainty_plots,
    _contour_plots,
    _load_data,
)
from .units import *
from ._chemicals import *
from ._contour_plots import *
from .systems import *
from ._tea import *
from .utils import *
from ._evaluation import *
from ._uncertainty_plots import *
from ._contour_plots import *
from ._load_data import *
from ._feature_mockups import *

__all__ = (
    *units.__all__,
    *_chemicals.__all__,
    *systems.__all__,
    *_tea.__all__,
    *utils.__all__,
    *_evaluation.__all__,
    *_uncertainty_plots.__all__,
    *_contour_plots.__all__,
    *_load_data.__all__,
    'sys',
    'tea', 
    'flowsheet',
)
from biorefineries.cane import Biorefinery

disable_derivative = Biorefinery.disable_derivative    
enable_derivative = Biorefinery.enable_derivative
cache = Biorefinery.cache

def load(*args, **kwargs):
    br = Biorefinery(*args, **kwargs)
    dct = globals()
    dct.update(br.__dict__)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
#                     Yalin Li <mailto.yalin.li@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import (utils,
               units,
               process_settings,
               chemicals,
               systems,
               tea,
               biorefinery,
)

__all__ = [*utils.__all__,
           *units.__all__,
           *process_settings.__all__,
           *chemicals.__all__,
           *systems.__all__,
           *tea.__all__,
           *biorefinery.__all__,
]

from .utils import *
from .units import *
from .process_settings import *
from .chemicals import *
from .systems import *
from .tea import *
from .biorefinery import *

def load(*args, **kwargs):
    br = Biorefinery(*args, **kwargs)
    globals().update(br.__dict__)
    globals().update({
        'biorefinery': br,
        'system': br.system,
        'tea': br.TEA,
    })



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
#                     Yalin Li <mailto.yalin.li@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
.. contents:: :local:

.. autofunction:: biorefineries.corn.systems.create_system

"""
import biosteam as bst
from biorefineries.ethanol import create_ethanol_purification_system
from .process_settings import BiorefinerySettings
from . import units

__all__ = ('create_system',)

import warnings
import numpy as np
warnings.filterwarnings("error", category=np.ComplexWarning)

def create_system(flowsheet=None, biorefinery_settings=None):
    settings = biorefinery_settings or BiorefinerySettings()
    system_ID = settings.system_ID
    feedstock_ID = 'corn' if 'corn' in system_ID else 'feedstock'
    parameters = settings.process_parameters
    prices = settings.stream_prices
    GWP_CFs = settings.stream_GWP_CFs
    
    get_stream_kwargs = lambda key: {
        'price': prices.get(key, 0.),
        'characterization_factors': {'GWP': GWP_CFs.get(key, 0.)},
        'units': 'kg/hr',
        }
    
    ### Streams ###
    
    chemicals = bst.settings.get_chemicals()
    
    z_mass_corn = chemicals.kwarray(settings.feedstock_composition)

    F_mass_corn = settings.feedstock_hourly_mass_flow
    mass_corn = F_mass_corn * z_mass_corn

    feedstock = bst.Stream(feedstock_ID, flow=mass_corn, **get_stream_kwargs(feedstock_ID.capitalize()))
    feedstock.register_alias('feedstock')
    crude_oil = bst.Stream('crude_oil', **get_stream_kwargs('Crude oil'))
    ammonia = bst.Stream('ammonia', NH3=89.723, **get_stream_kwargs('Ammonia'))
    lime = bst.Stream('lime', CaO=53.609, **get_stream_kwargs('Lime'))

    aa_price = prices.get('Alpha Amylase') or prices.get('Enzyme', 0.)
    aa_CF = GWP_CFs.get('Alpha Amylase') or prices.get('Enzyme', 0.)
    aa_kwargs = {
        'price': aa_price,
        'characterization_factors': {'GWP': aa_CF},
        'units': 'kg/hr',
        }

    alpha_amylase = bst.Stream('alpha_amylase',
                               SolubleProtein=0.00082,
                               Water=0.99918,
                               **aa_kwargs)
    # F_mass for this (and other feedstock-related streams)
    # will be adjusted through specification
    alpha_amylase.F_mass = 32.467
    recycled_process_water = bst.Stream('recycled_process_water', 
                                        Water=1, 
                                        units='kg/hr')
    ga_price = prices.get('Gluco Amylase') or prices.get('Enzyme', 0.)
    ga_CF = GWP_CFs.get('Gluco Amylase') or prices.get('Enzyme', 0.)
    ga_kwargs = {
        'price': ga_price,
        'characterization_factors': {'GWP': ga_CF},
        'units': 'kg/hr',
        }

    gluco_amylase = bst.Stream('gluco_amylase', 
                               SolubleProtein=0.0011,
                               Water=0.9989,
                               **ga_kwargs)
    gluco_amylase.F_mass = 46.895
    sulfuric_acid = bst.Stream('sulfuric_acid', 
                               H2SO4=1.,
                               **get_stream_kwargs('Sulfuric acid'))
    sulfuric_acid.F_mass = 92.59300
    backwater = bst.Stream('backwater', Water=22484., units='kg/hr')
    yeast = bst.Stream('yeast', Yeast=3.045, Water=63.335, **get_stream_kwargs('Yeast'))
    scrubber_water = bst.Stream('scrubber_water', Water=20139.4417, units='kg/hr')
    ethanol = bst.Stream('ethanol', **get_stream_kwargs('Ethanol'))
    denaturant = bst.Stream('denaturant', **get_stream_kwargs('Denaturant'))
    high_pressure_steam = bst.HeatUtility.get_agent('high_pressure_steam')
    DDGS = bst.Stream('DDGS', **get_stream_kwargs('DDGS'))
    steam = bst.Stream('steam', Water=1, phase='g',
                       T=high_pressure_steam.T,
                       P=high_pressure_steam.P,
                       **get_stream_kwargs('Steam'))
    
    ### Process specifications ###
    
    def refresh_feed_specifications():
        F_mass_dry_corn = feedstock.F_mass - feedstock.imass['Water']
        lime.F_mass = F_mass_dry_corn * parameters['slurry_lime_loading'] 
        ammonia.F_mass = F_mass_dry_corn * parameters['slurry_ammonia_loading']
        alpha_amylase.F_mass = F_mass_dry_corn * parameters['liquefaction_alpha_amylase_loading']
        sulfuric_acid.F_mass = F_mass_dry_corn * parameters['saccharification_sulfuric_acid_loading']
        MH101._run()
    
    def update_scrubber_wash_water():
        scrubber_water.F_mass =  V409.ins[1].F_mass * parameters['scrubber_wash_water_over_vent']
        V409._run()
        
    ### Units ###
    
    MH101 = units.GrainHandling('MH101', feedstock)
    MH101.add_specification(refresh_feed_specifications)
    V102 = units.CornStorage('V102', MH101-0)
    MH103 = units.CleaningSystem('MH103', V102-0, split=0.997)
    M104 = units.HammerMill('M104', MH103-0)
    V105 = units.MilledCornSurgeTank('V105', M104-0)
    W106 = units.MilledCornHopper('W106', V105-0)
    V107 = units.MilledCornWeighTank('V107', W106-0)
    V301 = units.AlphaAmylaseTank('V301', alpha_amylase)
    P302 = bst.Pump('P302', V301-0)
    V303 = units.AmmoniaTank('V303', ammonia)
    P304 = bst.Pump('P304', V303-0)
    V305 = units.LimeHopper('V305', lime)
    V307 = units.SlurryMixTank('V307', (V107-0, P302-0, P304-0, V305-0, recycled_process_water, backwater))
    slurry_solids_content = parameters['slurry_solids_content']
    @V307.add_specification(run=True)
    def correct_recycle_dilution_water():
        F_mass_dry_corn = feedstock.F_mass - feedstock.imass['Water']
        F_mass_others = lime.F_mass + ammonia.F_mass + alpha_amylase.F_mass + sulfuric_acid.F_mass + gluco_amylase.F_mass
        recycled_process_water.F_mass = F_mass_dry_corn * (1. - slurry_solids_content) / slurry_solids_content - F_mass_others
    
    P311 = bst.Pump('P311', V307-0, P=2e6)
    E312 = bst.HXprocess('E312', (P311-0, None), U=0.56783, ft=1.0)
    E313 = units.JetCooker('E313', (E312-0, steam))
    V314 = units.CookedSlurrySurgeTank('V314', E313-0)
    P308 = bst.Pump('P308', V314-0)
    P308-0-1-E312
    # E315 = bst.HXutility('E315', E312-1, U=0.9937, ft=1.0, T= + 273.15)
    
    HX101 = bst.HXutility('HX101', E312-1, U=1.5, T=87 + 273.15, ft=1.0)
    V310 = units.Liquefaction('V310', HX101-0)
    E316 = bst.HXprocess('E316', (V310-0, None), U=0.85174, ft=1.0, dT=5)
    
    V317 = units.GlucoAmylaseTank('V317', gluco_amylase)
    P318 = bst.Pump('P318', V317-0)
    V319 = units.SulfuricAcidTank('V319', sulfuric_acid)
    P320 = bst.Pump('P320', V319-0)
    V321 = units.Saccharification('V321', (P318-0, P320-0, E316-0))
    P322 = bst.Pump('P322', V321-0)
    E401 = bst.HXprocess('E401', (P322-0, None),
                         phase0='l', phase1='l',
                         U=0.99370, ft=0.85)
    E402 = bst.HXutility('E402', E401-0, ft=0.95, U=0.9937, T=32.2 + 273.15)
    V403 = units.YeastTank('V403', yeast)
    P404 = bst.Pump('P404', V403-0)
    V405 = units.SSF('V405', (E402-0, P404-0), outs=('CO2', ''), V=1.9e3)
    
    @V405.add_specification(run=True, impacted_units=[V405])
    def correct_saccharification_feed_flows():
        mash = V405.ins[0]
        mash_flow = mash.F_mass
        mash_dry_flow = mash_flow - mash.imass['Water']
        yeast.F_mass = parameters['yeast_loading'] * mash_flow
        gluco_amylase.F_mass = parameters['saccharification_gluco_amylase_loading'] * mash_dry_flow
    
    P406 = bst.Pump('P406', V405-1)
    P407 = bst.Pump('P407', E401-1)
    P407-0-1-E316
    V412 = bst.StorageTank('V412', E316-1, tau=4)
    PX = bst.Pump('PX', V412-0)
    V409 = bst.VentScrubber('V409', (scrubber_water, V405-0), gas=('CO2', 'O2'))
    V409.add_specification(update_scrubber_wash_water)
    P410 = bst.Pump('P410', V409-1)
    MX = bst.Mixer('MX', P406-0)
    MX-0-1-E401
    
    ethanol_purification_sys = create_ethanol_purification_system(
        ins=[PX-0, denaturant],
        outs=ethanol,
        beer_column_heat_integration=True,
        mockup=True,
        IDs={
            'Beer column': 'T501',
            'Beer column heat exchange': 'E413',
            'Beer column bottoms product pump': 'P502',
            'Recycle mixer': 'MX3',
            'Distillation': 'T503_T507',
            'Heat exchanger to superheat vapor to molecular sieves': 'HX500',
            'Molecular sieves': 'X504',
            'Ethanol condenser': 'HX501',
            'Ethanol day tank': 'V511', 
            'Ethanol day tank pump': 'P512',
            'Denaturant storage': 'V509', 
            'Denaturant pump': 'P510', 
            'Ethanol-denaturant mixer': 'MX4',
            'Product tank': 'V513',
            'Distillation bottoms product pump': 'P508',
        }
    )
    f = flowsheet or bst.main_flowsheet
    u = f.unit
    u.X504.approx_duty = False
    u.T501.Rmin = 0.0001
    u.T501.k = 1.05
    u.T503_T507.k = 1.05
    u.T501.P = 101325
    u.T503_T507.P = 101325
    u.E413.U = 0.79496
    u.E413.ft = 1.0
    u.MX4.denaturant_fraction = 0.04345
    V601 = bst.MixTank('V601', u.E413-1)
    P602 = bst.Pump('P602', V601-0)
    C603 = units.DDGSCentrifuge('C603', P602-0,
        split=dict(
            Water=0.8285,
            Ethanol=0.8285,
            Yeast=0.2734,
            Lipid=0.5, # 0.331 originally
            H2SO4=0.8285,
            Starch=0.08,
            Fiber=0.5328,
            SolubleProtein=0.8285,
            InsolubleProtein=0.08)
    )
    MH604 = units.WetDDGSConveyor('MH604', C603-1)
    S1 = bst.Splitter('S1', C603-0, (backwater, ''), split=0.208)
    V605 = bst.MixTank('V605', S1-1)
    P606 = bst.Pump('P606', V605-0)
    Ev607 = bst.MultiEffectEvaporator('Ev607',
        ins=P606-0,
        P=(101325, 69682, 47057, 30953),
        V=0.90
    ) 
    C603_2 = bst.LiquidsSplitCentrifuge('C603_2', Ev607-0, (crude_oil, ''), split={'Oil':0.99})
    
    MX5 = bst.Mixer('MX5', (Ev607-1, P410-0, u.P508-0))
    MX6 = bst.Mixer('MX6', (C603_2-1, MH604-0))
    D610 = bst.DrumDryer('D610', (MX6-0, 'dryer_air', 'natural_gas'), moisture_content=0.10, split=dict(Ethanol=1.0))
    X611 = bst.ThermalOxidizer('X611', (D610-1, 'oxidizer_air', ''))
    MH612 = units.DDGSHandling('MH612', D610-0, DDGS)
    T608 = bst.facilities.ProcessWaterCenter(
        'T608', 
        (MX5-0, 'makeup_water'), 
        ('process_water', 'wastewater'),
        None,
        (),
        (recycled_process_water,)
    )
    other_facilities = units.PlantAir_CIP_WasteWater_Facilities('other_facilities', feedstock)
    # HXN = bst.HeatExchangerNetwork('HXN', 
    #     units=lambda: [u.Ev607.evaporators[0], u.T503_T507.condenser]
    # )
    
    other_unit_parameters = settings.other_unit_parameters
    for ID, params in other_unit_parameters.items():
        attr, val = params.items()
        setattr(f.unit.search(ID), attr, val)
        
    globals().update(f.unit.data)
        
    return f.create_system(system_ID)
    
    # System = bst.System
    
    # return System('corn_sys',
    #     [MH101,
    #      V102,
    #      MH103,
    #      M104,
    #      V105,
    #      W106,
    #      V107,
    #      V303,
    #      P304,
    #      V305,
    #      V301,
    #      P302,
    #      V317,
    #      P318,
    #      V319,
    #      P320,
    #      V403,
    #      P404,
    #      System('SYS5',
    #         [V307,
    #          P311,
    #          System('SYS1',
    #             [E312,
    #              E313,
    #              V314,
    #              P308],
    #             recycle=P308-0),
    #          HX101,
    #          V310,
    #          System('SYS3',
    #             [E316,
    #              V321,
    #              P322,
    #              System('SYS2',
    #                 [E401,
    #                  E402,
    #                  V405,
    #                  P406],
    #                 recycle=P406-0),
    #              P407],
    #             recycle=P407-0),
    #          V412,
    #          System('SYS4',
    #             [E413,
    #              MX2,
    #              P411,
    #              P301,
    #              T501,
    #              P502],
    #             recycle=P502-0),
    #          V601,
    #          P602,
    #          C603,
    #          S1],
    #         recycle=S1-0),
    #      E408,
    #      E408_2,
    #      System('SYS6',
    #         [MX3,
    #          T503_T507,
    #          HX500,
    #          X504],
    #         recycle=X504-0),
    #      HX501,
    #      V511,
    #      P512,
    #      V509,
    #      P510,
    #      MX4,
    #      V513,
    #      V605,
    #      P606,
    #      Ev607,
    #      C603_2,
    #      MX6,
    #      D610,
    #      X611,
    #      MH612,
    #      P508,
    #      MX5,
    #      MX1,
    #      V409,
    #      P410,
    #      MH604],
    #     facilities=[other_facilities,
    #      T608,
    #      HXN])
                        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
"""
import numpy as np
import biosteam as bst
import flexsolve as flx
from .. import streams as s
from biosteam import SystemFactory

__all__ = (
    'create_sugar_crystallization_system',
)

@SystemFactory(
    ID='sugar_crystallization_sys',
    ins=[s.screened_juice, s.lime, s.H3PO4, s.polymer],
    outs=[s.sugar, s.molasses]
)
def create_sugar_crystallization_system(ins, outs):
    # TODO: Add conveyors, storage tanks, packing, sugar remelter
    # https://www.researchgate.net/profile/Maciej-Starzak/publication/311206128_Mass_and_Energy_Balance_Modelling_of_a_Sugar_Mill_A_comparison_of_MATLABR_and_SUGARS_simulations/links/583f240308ae2d217557dcd8/Mass-and-Energy-Balance-Modelling-of-a-Sugar-Mill-A-comparison-of-MATLABR-and-SUGARS-simulations.pdf?origin=publication_detail
    # https://www3.epa.gov/ttn/chief/ap42/ch09/final/c9s10-1a.pdf
    # http://sugartech.co.za/verticalcrystalliser/index.php
    screened_juice, lime, H3PO4, polymer = ins
    sugar, molasses, = outs
    
    if 'Sugar' not in sugar.chemicals:
        sugar.chemicals.define_group('Sugar', ('Glucose', 'Sucrose'))
    
    # Concentrate sugars
    P1 = bst.Pump('P1', ins=screened_juice, P=101325)
    
    MEE = bst.MultiEffectEvaporator('MEE', P1-0,
        P=(101325, 69682, 47057, 30953, 19781),
        V_definition='First-effect',
        V=0.3
    ) # fraction evaporated
    MEE.brix = 95
    
    def get_brix():
        effluent = MEE.outs[0]
        water = effluent.imass['Water']
        if water < 0.0001: water = 0.0001
        return 100 * effluent.imass['Sugar'] / water
    
    def brix_objective(V):
        MEE.V = V
        MEE._run()
        return MEE.brix - get_brix()
    
    @MEE.add_specification(run=False)
    def adjust_glucose_concentration():
        V_guess = MEE.V
        MEE.V = flx.IQ_interpolation(
            brix_objective, 0., 1., x=V_guess, ytol=1e-5
        )
    
    # Mix in flocculant
    T1 = bst.MixTank('T1', (MEE-0, lime, H3PO4, polymer))
    T1.tau = 0.10
    
    @T1.add_specification(run=True)
    def correct_flows():
        F_mass = T1.ins[0].F_mass
        # correct lime and phosphoric acid
        lime.imass['CaO', 'Water'] = 0.1 * F_mass * np.array([0.046, 0.954])
        H3PO4.imass['H3PO4', 'Water'] = 0.025 * F_mass
    
    # Separate residual solids
    C1 = bst.Clarifier('C1', T1-0, 
                           split=dict(Ash=0,
                                      Cellulose=0,
                                      Flocculant=1,
                                      Glucose=1,
                                      Hemicellulose=0,
                                      Lignin=0,
                                      CaO=1,
                                      H3PO4=1,
                                      Sucrose=1,
                                      Water=0.99))
    P2 = bst.Pump('P2', C1-0, P=101325)
    M1 = bst.Mixer('M1', (P2-0, '', ''))
    E1 = bst.Flash('E1', M1-0, V=0.5, P=15000)
    
    def get_purity(flash):
        effluent = flash.outs[1]
        return effluent.imass['Sugar'] / effluent.F_mass
    
    def purity_objective(V, flash):
        flash.V = V
        flash._run()
        return flash.purity - get_purity(flash)
    
    def adjust_purity(flash):
        V_guess = flash.V
        y0 = purity_objective(0., flash)
        if y0 < 0.: return
        y1 = purity_objective(1., flash)
        if y1 > 0.: return
        flash.V = flx.IQ_interpolation(
            purity_objective, 0., 1., y0, y1, x=V_guess, ytol=1e-5,
            args=(flash,),
        )
    
    E1.add_specification(adjust_purity, run=False, args=(E1,))
    E1.purity = 0.8623
    BC1 = bst.BatchCrystallizer('BC1', E1-1, tau=8, V=3785, T=55 + 273.15)
    
    def get_split(molasses_flow, molasses_purity, crystal_flow, crystal_purity):
        s_crystal = crystal_flow * crystal_purity
        s_molasses = molasses_flow * molasses_purity
        s_split = s_crystal / (s_crystal + s_molasses)
        o_crystal = crystal_flow * (100 - crystal_purity)
        o_molasses = molasses_flow * (100 - molasses_purity)
        o_split = o_crystal / (o_crystal + o_molasses)
        return dict(
            Water=o_split,
            H3PO4=o_split,
            CaO=o_split,
            Sugar=s_split,
        )
    
    C2 = bst.SolidsCentrifuge('C2', 
        BC1-0, 
        split=get_split(19.53, 62.91, 29.68, 98.61),
        moisture_content=None,
    )
    
    bst.StorageTank('S1', C2-0, sugar, tau=27 * 7)
    
    def correct_wash_water(mixer):
        mixer.ins[1].imass['Water'] = mixer.ins[0].imass['Sugar']
    
    M2 = bst.Mixer('M2', (C2-1, ''))
    M2.add_specification(correct_wash_water, run=True, args=(M2,))
    P3 = bst.Pump('P3', M2-0, P=101325)
    E2 = bst.Flash('E2', P3-0, V=0.5, P=15000)
    E2.add_specification(adjust_purity, run=False, args=(E2,))
    E2.purity = 0.6291
    BC2 = bst.BatchCrystallizer('BC2', E2-1, tau=24, V=3785, T=50 + 273.15)
    C3 = bst.SolidsCentrifuge('C3', 
        BC2-0, (2-M1, ''),
        split=get_split(4.34, 33.88, 3.15, 96.49),
        moisture_content=None,
    )
    M3 = bst.Mixer('M3', (C3-1, ''))
    M3.add_specification(correct_wash_water, run=True, args=(M3,))
    P4 = bst.Pump('P4', M3-0, P=101325)
    E3 = bst.Flash('E3', P4-0, V=0.5, P=15000)
    E3.add_specification(adjust_purity, run=False, args=(E3,))
    E3.purity = 0.5450
    BC3 = bst.BatchCrystallizer('BC3', E3-1, tau=40, V=3785, T=45 + 273.15)
    C4 = bst.SolidsCentrifuge('C4', 
        BC3-0, (1-M1, ''),
        split=get_split(9.04, 32.88, 4.48, 93.84),
        moisture_content=None,
    )
    bst.StorageTank('S2', C4-1, molasses, tau=24 * 7)



# -*- coding: utf-8 -*-
"""
.. autofunction:: biorefineries.cellulosic.systems.cellulosic_ethanol.create_cellulosic_ethanol_system

"""

import biosteam as bst
import thermosteam as tmo
from thermosteam import Stream
from .pretreatment import create_dilute_acid_pretreatment_system
from .fermentation import create_cellulosic_fermentation_system
from biorefineries.ethanol import create_ethanol_purification_system
from biorefineries.cellulosic import streams as s
from biorefineries.cellulosic import units


__all__ = (
    'create_cellulosic_ethanol_system',
    'create_facilities',
)

def create_facilities(
        solids_to_boiler, 
        gas_to_boiler,
        process_water_streams,
        feedstock,
        RO_water='',
        recycle_process_water='',
        blowdown_to_wastewater=None,
        BT_area=None,
        area=None,
        BTkw=None
    ):
    
    BT = bst.facilities.BoilerTurbogenerator(BT_area or area or 'BT',
                                             **(BTkw or {}),
                                             ins=(solids_to_boiler,
                                                  gas_to_boiler, 
                                                  'boiler_makeup_water',
                                                  'natural_gas',
                                                  'FGD_lime',
                                                  'boilerchems'))
    
    bst.facilities.ChilledWaterPackage(area or 'CWP')
    CT = bst.facilities.CoolingTower(area or 'CT')
    
    process_water_streams = (*process_water_streams,
                              BT-1, CT-1)
            
    makeup_water = Stream('makeup_water', price=0.00025353160151260924)
    
    bst.facilities.ProcessWaterCenter(area or 'PWC',
        (RO_water, '', recycle_process_water, makeup_water),
        (),
        None,
        (BT-1, CT-1),
        process_water_streams
    )
    CIP = Stream('CIP', Water=126, units='kg/hr')
    CIP_package = bst.facilities.CIPpackage(area or 'CIP_package', CIP)
    CIP_package.CIP_over_feedstock = 0.00121
    @CIP_package.add_specification(run=True)
    def adjust_CIP(): CIP.imass['Water'] = feedstock.F_mass * CIP_package.CIP_over_feedstock
    
    plant_air = Stream('plant_air', N2=83333, units='kg/hr')
    ADP = bst.facilities.AirDistributionPackage(area or 'ADP', plant_air)
    ADP.plant_air_over_feedstock = 0.8
    @ADP.add_specification(run=True)
    def adjust_plant_air(): plant_air.imass['N2'] = feedstock.F_mass * ADP.plant_air_over_feedstock
        
    fire_water = Stream('fire_water', Water=8343, units='kg/hr')
    FT = bst.FireWaterTank(area or 'FT', fire_water)
    FT.fire_water_over_feedstock = 0.08
    @FT.add_specification(run=True)
    def adjust_fire_water(): fire_water.imass['Water'] = feedstock.F_mass * FT.fire_water_over_feedstock
    
    ### Complete system
    if blowdown_to_wastewater:
        bst.BlowdownMixer(area or 'blowdown_mixer', (BT-1, CT-1), blowdown_to_wastewater)

@bst.SystemFactory(
    ID='cornstover_sys',
    ins=[*create_dilute_acid_pretreatment_system.ins,
          s.denaturant],
    outs=[s.ethanol],
)
def create_cellulosic_ethanol_system(
        ins, outs,
        include_blowdown_recycle=None,
        WWT_kwargs=None,
    ):
    feedstock, sulfuric_acid, ammonia, denaturant = ins
    ethanol, = outs
    U101 = units.FeedStockHandling('U101', feedstock)
    U101.cost_items['System'].cost = 0.
    pretreatment_sys = create_dilute_acid_pretreatment_system(
        ins=[U101-0, sulfuric_acid, ammonia],
        mockup=True
    )
    fermentation_sys = create_cellulosic_fermentation_system(
        ins=pretreatment_sys-0,
        mockup=True,
    )
    ethanol_purification_sys, udct = create_ethanol_purification_system(
        ins=[fermentation_sys-1, denaturant],
        outs=[ethanol],
        udct=True,
        IDs={'Beer pump': 'P401',
             'Beer column heat exchange': 'H401',
             'Beer column': 'D402',
             'Beer column bottoms product pump': 'P402',
             'Distillation': 'D403',
             'Distillation bottoms product pump': 'P403',
             'Ethanol-denaturant mixer': 'M701',
             'Recycle mixer': 'M402',
             'Heat exchanger to superheat vapor to molecular sieves': 'H402',
             'Molecular sieves': 'U401',
             'Ethanol condenser': 'H403',
             'Ethanol day tank': 'T701', 
             'Ethanol day tank pump': 'P701',
             'Denaturant storage': 'T702', 
             'Denaturant pump': 'P702', 
             'Product tank': 'T703'},
        mockup=True,
    )
    udct['H401'].dT = 10
    udct['D402'].k = 1.4
    udct['D403'].k = 1.4
    ethanol, stillage, recycle_process_water = ethanol_purification_sys.outs
    recycled_water = tmo.Stream(Water=1,
                                T=47+273.15,
                                P=3.9*101325,
                                units='kg/hr')
    S401 = bst.PressureFilter('S401', (stillage, recycled_water))

    bst.create_all_facilities(
        feedstock, 
        blowdown_recycle=include_blowdown_recycle,
        WWT_kwargs=WWT_kwargs,
        HXN=False,
        recycle_process_water_streams=[recycle_process_water],
    )



# -*- coding: utf-8 -*-
"""
Created on Thu Dec 15 20:56:00 2022

@author: yrc2
"""
from biorefineries.cellulosic import systems

__all__ = [*systems.__all__, 'create_system']

from biorefineries.cellulosic.systems import *

create_system = create_cellulosic_ethanol_system


# -*- coding: utf-8 -*-
"""
.. contents:: :local:

Reactors
--------
.. autoclass:: biorefineries.cellulosic.units.PretreatmentReactorSystem
.. autoclass:: biorefineries.cellulosic.units.SeedTrain
.. autoclass:: biorefineries.cellulosic.units.ContinuousPresaccharification
.. autoclass:: biorefineries.cellulosic.units.Saccharification
.. autoclass:: biorefineries.cellulosic.units.CoFermentation
.. autoclass:: biorefineries.cellulosic.units.SaccharificationAndCoFermentation
.. autoclass:: biorefineries.cellulosic.units.SimultaneousSaccharificationAndCoFermentation

Separations
-----------
.. autoclass:: biorefineries.cellulosic.units.ReverseOsmosis
.. autoclass:: biorefineries.cellulosic.units.Nanofilter
.. autoclass:: biorefineries.cellulosic.units.HydrolyzateSolidLiquidSeparator
.. autoclass:: biorefineries.cellulosic.units.PretreatmentFlash

Feedstock handling
------------------
.. autoclass:: biorefineries.cellulosic.units.FeedStockHandling

Tanks
-----
.. autoclass:: biorefineries.cellulosic.units.AmmoniaStorageTank
.. autoclass:: biorefineries.cellulosic.units.AmmoniaReacidificationTank
.. autoclass:: biorefineries.cellulosic.units.BeerTank
.. autoclass:: biorefineries.cellulosic.units.CSLStorageTank
.. autoclass:: biorefineries.cellulosic.units.DAPStorageTank
.. autoclass:: biorefineries.cellulosic.units.AmmoniaAdditionTank
.. autoclass:: biorefineries.cellulosic.units.AmmoniaMixer
.. autoclass:: biorefineries.cellulosic.units.EnzymeHydrolysateMixer
.. autoclass:: biorefineries.cellulosic.units.SulfuricAcidMixer
.. autoclass:: biorefineries.cellulosic.units.SeedHoldTank
.. autoclass:: biorefineries.cellulosic.units.SulfuricAcidStorageTank
.. autoclass:: biorefineries.cellulosic.units.SulfuricAcidTank
.. autoclass:: biorefineries.cellulosic.units.OligomerConversionTank

Heat exchange
-------------
.. autoclass:: biorefineries.cellulosic.units.HydrolysateHeatExchanger
.. autoclass:: biorefineries.cellulosic.units.PretreatmentWasteHeater
.. autoclass:: biorefineries.cellulosic.units.WasteVaporCondenser

Pumps
-----
.. autoclass:: biorefineries.cellulosic.units.HydrolysatePump
.. autoclass:: biorefineries.cellulosic.units.BlowdownDischargePump
.. autoclass:: biorefineries.cellulosic.units.HydrolyzatePump
.. autoclass:: biorefineries.cellulosic.units.ReacidifiedHydrolyzatePump

"""
from biosteam.units.design_tools.geometry import cylinder_diameter_from_volume
from thermosteam import MultiStream
from biosteam import Unit
from biosteam.units.decorators import cost
from biosteam.units.design_tools import size_batch
import thermosteam as tmo
import biosteam as bst
import numpy as np

__all__ = (
    'PretreatmentReactorSystem',
    'SeedTrain',
    'ContinuousPresaccharification',
    'Saccharification',
    'CoFermentation',
    'SaccharificationAndCoFermentation',
    'SimultaneousSaccharificationAndCoFermentation',
    'ReverseOsmosis',
    'Nanofilter',
    'HydrolysatePump',
    'AmmoniaStorageTank',
    'AmmoniaReacidificationTank',
    'BeerTank',
    'BlowdownDischargePump',
    'CSLStorageTank',
    'DAPStorageTank',
    'FeedStockHandling',
    'PretreatmentFlash',
    'HydrolysateHeatExchanger',
    'PretreatmentWasteHeater',
    'WasteVaporCondenser',
    'HydrolyzatePump',
    'HydrolyzateSolidLiquidSeparator',
    'AmmoniaAdditionTank',
    'AmmoniaMixer',
    'EnzymeHydrolysateMixer',
    'SulfuricAcidMixer',
    'OligomerConversionTank',
    'ReacidifiedHydrolyzatePump',
    'SeedHoldTank',
    'SulfuricAcidStorageTank',
    'SulfuricAcidTank',
)

Rxn = tmo.reaction.Reaction
ParallelRxn = tmo.reaction.ParallelReaction

# %% Constants

_gal2m3 = 0.003785
_gpm2m3hr = 0.227124
# _m3hr2gpm = 4.40287
_hp2kW = 0.7457
_Gcal2kJ = 4184e3

# %% Pretreatment

@cost('Dry flow rate', 'Pretreatment reactor system', units='kg/hr',
      S=83333, CE=522, cost=19812400 * 0.993, n=0.6, kW=4578, BM=1.5)
class PretreatmentReactorSystem(bst.units.design_tools.PressureVessel, Unit):
    _N_ins = 1
    _N_outs = 2
    _graphics = bst.Flash._graphics
    _units = {'Residence time': 'hr',
              'Reactor volume': 'm3'}
    
    def __init__(self, ID='', ins=None, outs=(), T=130+273.15, thermo=None, 
                 tau=0.166, V_wf=0.8, length_to_diameter=2, 
                 vessel_material='Stainless steel 316', 
                 vessel_type='Horizontal',
                 reactions=None,
                 run_vle=True):
        Unit.__init__(self, ID, ins, outs, thermo)
        self._load_components()
        vapor, liquid = self.outs
        vapor.phase = 'g'
        self.T = T
        chemicals = self.chemicals
        if reactions is None:
            self.reactions = ParallelRxn([
        #            Reaction definition                 Reactant    Conversion
        Rxn('Glucan + H2O -> Glucose',                   'Glucan',   0.0990, chemicals),
        Rxn('Glucan + H2O -> GlucoseOligomer',           'Glucan',   0.0030, chemicals),
        Rxn('Glucan -> HMF + 2 H2O',                     'Glucan',   0.0030, chemicals),
        Rxn('Galactan + H2O -> GalactoseOligomer',       'Galactan', 0.0240, chemicals),
        Rxn('Galactan -> HMF + 2 H2O',                   'Galactan', 0.0030, chemicals),
        Rxn('Mannan + H2O -> MannoseOligomer',           'Mannan',   0.0030, chemicals),
        Rxn('Mannan -> HMF + 2 H2O',                     'Mannan',   0.0030, chemicals),
        Rxn('Sucrose -> HMF + Glucose + 2H2O',           'Sucrose',  1.0000, chemicals),
        Rxn('Xylan + H2O -> Xylose',                     'Xylan',    0.9000, chemicals),
        Rxn('Xylan + H2O -> XyloseOligomer',             'Xylan',    0.0240, chemicals),
        Rxn('Xylan -> Furfural + 2 H2O',                 'Xylan',    0.0500, chemicals),
        Rxn('Arabinan + H2O -> Arabinose',               'Arabinan', 0.9000, chemicals),
        Rxn('Arabinan + H2O -> ArabinoseOligomer',       'Arabinan', 0.0240, chemicals),
        Rxn('Arabinan -> Furfural + 2 H2O',              'Arabinan', 0.0050, chemicals),
        Rxn('Acetate -> AceticAcid',                     'Acetate',  1.0000, chemicals),
        Rxn('Lignin -> SolubleLignin',                   'Lignin',   0.0500, chemicals)
            ])
            self.glucan_to_glucose = self.reactions[0]
            self.xylan_to_xylose = self.reactions[8]
            self.glucose_to_byproducts = self.reactions[1:3]
            self.xylose_to_byproducts = self.reactions[9:12]
        else:
            self.reactions = reactions
        self.tau = tau
        self.V_wf = V_wf
        self.length_to_diameter = length_to_diameter
        self.vessel_material = vessel_material
        self.vessel_type = vessel_type
        self.run_vle = run_vle
        
    def _load_components(self):
        thermo = self.thermo
        self._multistream = MultiStream(None, thermo=thermo)
    
    def _run(self):
        feed = self.ins[0]
        vapor, liquid = self.outs
        liquid.copy_like(feed)
        self.reactions.adiabatic_reaction(liquid) 
        if self.T:
            if self.run_vle:
                ms = self._multistream
                ms.copy_like(liquid)
                ms.vle(T=self.T, H=ms.H)
                vapor.mol[:] = ms.imol['g']
                liquid.mol[:] = ms.imol['l']
                vapor.T = liquid.T = ms.T
                vapor.P = liquid.P = ms.P
            else:
                liquid.T = self.T

    def _design(self):
        Design = self.design_results
        ins_F_vol = self.F_vol_in
        V_reactor = ins_F_vol * self.tau / self.V_wf
        P = self.outs[0].P * 0.000145038 # Pa to psi
        length_to_diameter = self.length_to_diameter
        D = cylinder_diameter_from_volume(V_reactor, self.length_to_diameter)
        D *= 3.28084 # convert from m to ft
        L = D * length_to_diameter
        Design['Residence time'] = self.tau
        Design['Reactor volume'] = V_reactor
        Design.update(self._vessel_design(float(P), float(D), float(L)))
        self._decorated_design()
            
    def _cost(self):
        Design = self.design_results
        self.baseline_purchase_costs.update(
            self._vessel_purchase_cost(
                Design['Weight'], Design['Diameter'], Design['Length']
            )
        )
        self._decorated_cost()
    

@cost('Flow rate', 'Pumps',
      S=43149, CE=522, cost=24800, n=0.8, kW=40, BM=2.3)
@cost('Stage #1 reactor volume', 'Stage #1 reactors',
      cost=37700, S=20*_gal2m3, CE=522, n=0.7, BM=1.8)
@cost('Stage #2 reactor volume', 'Stage #2 reactors',
      cost=58300, S=200*_gal2m3, CE=522, n=0.7, BM=1.8)
@cost('Stage #3 reactor volume', 'Stage #3 reactors',
      cost=78800, S=2e3*_gal2m3, CE=522, n=0.7, BM=1.8)
@cost('Stage #4 reactor volume', 'Stage #4 reactors',
      cost=176e3, S=20e3*_gal2m3, CE=522, n=0.7, BM=1.8)
@cost('Stage #4 reactor volume', 'Stage #4 agitators',
      cost=26e3/2, S=20e3*_gal2m3, kW=7.5, CE=522, n=0.5, BM=1.5)
@cost('Stage #5 reactor volume', 'Stage #5 reactors',
      cost=590e3, S=200e3*_gal2m3, CE=522, n=0.7, BM=1.8)
@cost('Stage #5 reactor volume', 'Stage #5 agitators',
      cost=43e3/2, S=200e3*_gal2m3, kW=10, CE=522, n=0.5, BM=1.5)
class SeedTrain(Unit):
    _N_ins = 1
    _N_outs = 2
    _ins_size_is_fixed = False
    
    _units= {'Flow rate': 'kg/hr',
             'Stage #1 reactor volume': 'm3',
             'Stage #2 reactor volume': 'm3',
             'Stage #3 reactor volume': 'm3',
             'Stage #4 reactor volume': 'm3',
             'Stage #5 reactor volume': 'm3'}
    
    @property
    def N_stages(self): 
        """Number of stages in series."""
        return 5
    
    #: Number of parallel seed trains
    N_trains = 2
    
    #: Cycle time for each batch (hr)
    tau_batch = 24
    
    @property
    def tau_turnover(self):
        """Turnover time (hr) calculated by batch time divided by number of trains."""
        return self.tau_batch/self.N_trains
    
    #: Operating temperature (K)
    T = 32+273.15
    
    # #: wt % media (e.g. corn steep liquor) in each stage 
    # media_loading = 0.50
    
    # #: Diammonium phosphate loading in g/L of fermentation broth
    # DAP = 0.67 
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, 
                 reactions=None, saccharification=None):
        Unit.__init__(self, ID, ins, outs, thermo)
        chemicals = self.chemicals
        if reactions is None:
            self.reactions = ParallelRxn([
        #   Reaction definition                             Reactant    Conversion
        Rxn('Glucose -> 2 Ethanol + 2 CO2',                 'Glucose',   0.9000, chemicals),
        Rxn('Glucose + 0.047 CSL + 0.018 DAP -> 6 Z_mobilis + 2.4 H2O',
                                                            'Glucose',   0.0400, chemicals),
        Rxn('Glucose + 2 H2O -> 2 Glycerol + O2',           'Glucose',   0.0040, chemicals),
        Rxn('Glucose + 2 CO2 -> 2 SuccinicAcid + O2',       'Glucose',   0.0060, chemicals),
        Rxn('3 Xylose -> 5 Ethanol + 5 CO2',                'Xylose',    0.8000, chemicals),
        Rxn('Xylose + 0.039 CSL + 0.015 DAP -> 5 Z_mobilis + 2 H2O',
                                                            'Xylose',    0.0400, chemicals),
        Rxn('3 Xylose + 5 H2O -> 5 Glycerol + 2.5 O2',      'Xylose',    0.0030, chemicals),
        Rxn('Xylose + H2O -> Xylitol + 0.5 O2',             'Xylose',    0.0460, chemicals),
        Rxn('3 Xylose + 5 CO2 -> 5 SuccinicAcid + 2.5 O2',  'Xylose',    0.0090, chemicals)
            ])
            self.glucose_to_ethanol = self.reactions[0]
            self.xylose_to_ethanol = self.reactions[4]
            self.glucose_to_byproducts = self.reactions[1:4]
            self.xylose_to_byproducts = self.reactions[5:]
        else:
            self.reactions = reactions
        if callable(saccharification):
            self.saccharification = saccharification
        elif saccharification:
            self.saccharification = ParallelRxn([
                Rxn('Glucan -> GlucoseOligomer',          'Glucan',   0.0400, chemicals),
                Rxn('Glucan + 0.5 H2O -> 0.5 Cellobiose', 'Glucan',   0.0120, chemicals),
                Rxn('Glucan + H2O -> Glucose',            'Glucan',   0.9000, chemicals),
                Rxn('Cellobiose + H2O -> 2Glucose',       'Cellobiose',  1.0000, chemicals)]
            )
        else:
            self.saccharification = None
    
    _setup = Unit._setup
    
    def _run(self):
        vent, effluent= self.outs
        effluent.mix_from(self.ins, energy_balance=False)
        if self.saccharification:
            self.saccharification(effluent)
        self.reactions.force_reaction(effluent)
        effluent.empty_negative_flows()
        effluent.T = self.T
        vent.phase = 'g'
        vent.copy_flow(effluent, ('CO2', 'O2'), remove=True)

    def _design(self): 
        maxvol = self.outs[1].F_vol*self.tau_turnover
        vol = maxvol*10**-self.N_stages
        Design = self.design_results
        for i in range(1, self.N_stages+1):
            Design[f'Stage #{i} reactor volume'] = vol
            vol *= 10 
        Design['Flow rate'] = sum([i.F_mass for i in self.outs])
        self.add_heat_utility(self.Hnet, self.T)

    def _cost(self):
        N = self.N_trains
        D = self.design_results
        C = self.baseline_purchase_costs
        kW = 0
        for i, x in self.cost_items.items():
            S = D[x._basis]
            q = S/x.S
            C[i] = N*bst.CE/x.CE*x.cost*q**x.n
            kW += N*x.kW*q
        self.power_utility(kW)
        

# %% Saccharification and fermentation (consolidated bioprocess)

@cost('Flow rate', 'Transfer pumps', kW=58, S=352*_gpm2m3hr,
      cost=47200/5, CE=522, n=0.8, BM=2.3, N='N_transfer_pumps')
@cost('Tank volume', 'Tanks', cost=3840e3/8, S=250e3*_gal2m3, 
      CE=522, n=0.7, BM=2.0, N='N_tanks')
class ContinuousPresaccharification(Unit):
    _N_ins = 1
    _N_outs = 1
    
    #: Residence time of countinuous saccharification tanks (hr)
    tau_tank = 24
    
    #: Number of continuous saccharification tanks
    N_tanks = 8
    
    #: Number of transfer pumps
    N_transfer_pumps = 5
    
    #: Working volume fraction (filled tank to total tank volume)
    V_wf = 0.9
    
    _units = {'Flow rate': 'm3/hr',
              'Tank volume': 'm3'}
    
    def __init__(self, ID='', ins=None, outs=(), P=101325, reactions=None):
        Unit.__init__(self, ID, ins, outs)
        self.P = P
        self.reactions = reactions 
        
    def _run(self):
        inlet = self.ins[0]
        outlet = self.outs[0]
        outlet.copy_flow(inlet)
        outlet.T = inlet.T
        outlet.P = self.P
        if self.reactions: self.reactions.adiabatic_reaction(outlet)
        
    def _design(self):
        inlet = self.ins[0]
        v_0 = inlet.F_vol
        Design = self.design_results
        Design['Tank volume'] = v_0 * self.tau_tank / self.V_wf / self.N_tanks
        Design['Flow rate'] = v_0 / self.N_transfer_pumps


class Saccharification(bst.BatchBioreactor):
    _N_ins = 1
    _N_outs = 1
        
    #: Unload and clean up time (hr)
    tau_0 = 4
    
    #: Working volume fraction (filled tank to total tank volume)
    V_wf = 0.9
    
    _units = {'Flow rate': 'm3/hr',
              'Reactor volume': 'm3',
              'Reactor duty': 'kJ/hr'}
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, 
                 tau=72, N=None, V=3785.4118, T=48+273.15, P=101325,
                 Nmin=2, Nmax=36, reactions=None):
        bst.BatchBioreactor.__init__(self, ID, ins, outs, thermo, tau, N, V, T, P, Nmin, Nmax)
        chemicals = self.chemicals
        #: [ParallelReaction] Enzymatic hydrolysis reactions including from 
        #: downstream batch tank in co-fermentation.
        self.reactions = reactions or ParallelRxn([
            Rxn('Glucan -> GlucoseOligomer',          'Glucan',   0.0400, chemicals),
            Rxn('Glucan + 0.5 H2O -> 0.5 Cellobiose', 'Glucan',   0.0120, chemicals),
            Rxn('Glucan + H2O -> Glucose',            'Glucan',   0.9000, chemicals),
            Rxn('Cellobiose + H2O -> 2Glucose',       'Cellobiose',  1.0000, chemicals)]
        )
        
    _setup = Unit._setup
        
    @property
    def saccharification(self):
        return self.reactions
        
    @property
    def vent(self):
        return None
    @property
    def effluent(self):
        return self.outs[0]
        
    def _run(self):
        feed, = self.ins
        effluent, = self.outs
        effluent.copy_flow(feed)
        effluent.T = self.T
        effluent.P = self.P
        self.reactions(effluent)

class CoFermentation(bst.BatchBioreactor):
    _N_ins = 1
    _ins_size_is_fixed = False
        
    #: Unload and clean up time (hr)
    tau_0 = 4
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None,
                 tau=36, N=None, V=3785.4118, T=305.15, P=101325,
                 Nmin=2, Nmax=36, cofermentation=None, loss=None):
        bst.BatchBioreactor.__init__(self, ID, ins, outs, thermo, tau, N, V, T, P, Nmin, Nmax)
        self.P = P
        chemicals = self.chemicals
        self.loss = loss or ParallelRxn([
    #   Reaction definition               Reactant    Conversion
    Rxn('Glucose -> 2 LacticAcid',       'Glucose',   0.0300, chemicals),
    Rxn('3 Xylose -> 5 LacticAcid',      'Xylose',    0.0300, chemicals),
    Rxn('3 Arabinose -> 5 LacticAcid',   'Arabinose', 0.0300, chemicals),
    Rxn('Galactose -> 2 LacticAcid',     'Galactose', 0.0300, chemicals),
    Rxn('Mannose -> 2 LacticAcid',       'Mannose',   0.0300, chemicals),
        ])
        if cofermentation is None:
            self.cofermentation = ParallelRxn([
        #   Reaction definition                                          Reactant    Conversion
        Rxn('Glucose -> 2 Ethanol + 2 CO2',                             'Glucose',   0.9500, chemicals),
        Rxn('Glucose + 0.047 CSL + 0.018 DAP -> 6 Z_mobilis + 2.4 H2O', 'Glucose',   0.0200, chemicals),
        Rxn('Glucose + 2 H2O -> 2 Glycerol + O2',                       'Glucose',   0.0040, chemicals),
        Rxn('Glucose + 2 CO2 -> 2 SuccinicAcid + O2',                   'Glucose',   0.0060, chemicals),
        Rxn('3 Xylose -> 5 Ethanol + 5 CO2',                            'Xylose',    0.8500, chemicals),
        Rxn('Xylose + 0.039 CSL + 0.015 DAP -> 5 Z_mobilis + 2 H2O',
                                                                        'Xylose',    0.0190, chemicals),
        Rxn('3 Xylose + 5 H2O -> 5 Glycerol + 2.5 O2',                  'Xylose',    0.0030, chemicals),
        Rxn('Xylose + H2O -> Xylitol + 0.5 O2',                         'Xylose',    0.0460, chemicals),
        Rxn('3 Xylose + 5 CO2 -> 5 SuccinicAcid + 2.5 O2',              'Xylose',    0.0090, chemicals),
            ])
            self.glucose_to_ethanol = self.cofermentation[0]
            self.xylose_to_ethanol = self.cofermentation[4]
            self.glucose_to_byproducts = self.cofermentation[1:4]
            self.xylose_to_byproducts = self.cofermentation[5:]
        else:
            self.cofermentation = cofermentation
        
        if 'CSL' in self.chemicals:
            self.CSL_to_constituents = Rxn(
                'CSL -> 0.5 H2O + 0.25 LacticAcid + 0.25 Protein', 'CSL', 1.0000, chemicals, basis='wt',
            )
            self.CSL_to_constituents.basis = 'mol'
        else:
            self.CSL_to_constituents = None
        
        if all([i in self.chemicals for i in ('FFA', 'DAG', 'TAG', 'Glycerol')]):
            self.oil_reaction = self.lipid_reaction = ParallelRxn([
                Rxn('TAG + 3 Water -> 3FFA + Glycerol', 'TAG', 0.23, chemicals),
                Rxn('TAG + Water -> FFA + DAG', 'TAG', 0.02, chemicals)
            ])
        else:
            self.oil_reaction = self.lipid_reaction = None
        
    def _run(self):
        feeds = self.ins
        vent, effluent = self.outs
        vent.P = effluent.P = self.P
        vent.T = effluent.T = self.T
        vent.phase = 'g'
        effluent.mix_from(feeds, energy_balance=False)
        if self.loss: self.loss(effluent)
        if self.lipid_reaction: 
            self.lipid_reaction.force_reaction(effluent)
            effluent.empty_negative_flows()
        self.cofermentation(effluent)
        if self.CSL_to_constituents: self.CSL_to_constituents(effluent)
        vent.empty()
        vent.receive_vent(effluent, energy_balance=False)
        

@cost('Flow rate', 'Recirculation pumps', kW=30, S=340*_gpm2m3hr,
      cost=47200, n=0.8, BM=2.3, CE=522, N='N_reactors')
@cost('Reactor duty', 'Heat exchangers', CE=522, cost=23900,
      S=-5*_Gcal2kJ, n=0.7, BM=2.2, N='N_reactors') # Based on a similar heat exchanger
@cost('Batch duty', 'Fermentor batch cooler', CE=522, cost=86928,
      S=-5*_Gcal2kJ, n=0.7, BM=1.8) # Based on a similar heat exchanger
@cost('Reactor volume', 'Agitators', CE=522, cost=52500,
      S=1e6*_gal2m3, n=0.5, kW=90, BM=1.5, N='N_reactors')
@cost('Reactor volume', 'Reactors', CE=522, cost=844000,
      S=1e6*_gal2m3, n=0.5, BM=1.5, N='N_reactors')
class SaccharificationAndCoFermentation(Unit):
    _N_ins = 1
    _N_outs = 3
    _ins_size_is_fixed = False
        
    #: Saccharification temperature (K)
    T_saccharification = 48+273.15
    
    #: Fermentation temperature (K)
    T_fermentation = 32+273.15
    
    #: Saccharification time (hr)
    tau_saccharification = 60
    
    #: Co-Fermentation time (hr)
    tau_cofermentation = 36
    
    #: Unload and clean up time (hr)
    tau_0 = 4
    
    #: Working volume fraction (filled tank to total tank volume)
    V_wf = 0.9
    
    #: Number of reactors
    N_reactors = 12
    
    _units = {'Flow rate': 'm3/hr',
              'Reactor volume': 'm3',
              'Batch duty': 'kJ/hr',
              'Reactor duty': 'kJ/hr'}
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, P=101325, saccharification_split=0.1,
                 saccharification=None, loss=None, cofermentation=None):
        Unit.__init__(self, ID, ins, outs, thermo)
        self.P = P
        self.saccharification_split = saccharification_split
        chemicals = self.chemicals
        #: [ParallelReaction] Enzymatic hydrolysis reactions including from 
        #: downstream batch tank in co-fermentation.
        self.saccharification = saccharification or ParallelRxn([
            Rxn('Glucan -> GlucoseOligomer',          'Glucan',   0.0400, chemicals),
            Rxn('Glucan + 0.5 H2O -> 0.5 Cellobiose', 'Glucan',   0.0120, chemicals),
            Rxn('Glucan + H2O -> Glucose',            'Glucan',   0.9000, chemicals),
            Rxn('Cellobiose + H2O -> 2Glucose',       'Cellobiose',  1.0000, chemicals)]
        )
        self.loss = loss or ParallelRxn([
    #   Reaction definition               Reactant    Conversion
    Rxn('Glucose -> 2 LacticAcid',       'Glucose',   0.0300, chemicals),
    Rxn('3 Xylose -> 5 LacticAcid',      'Xylose',    0.0300, chemicals),
    Rxn('3 Arabinose -> 5 LacticAcid',   'Arabinose', 0.0300, chemicals),
    Rxn('Galactose -> 2 LacticAcid',     'Galactose', 0.0300, chemicals),
    Rxn('Mannose -> 2 LacticAcid',       'Mannose',   0.0300, chemicals),])
        self.cofermentation = cofermentation or ParallelRxn([
    #   Reaction definition                                          Reactant    Conversion
    Rxn('Glucose -> 2 Ethanol + 2 CO2',                             'Glucose',   0.9500, chemicals),
    Rxn('Glucose + 0.047 CSL + 0.018 DAP -> 6 Z_mobilis + 2.4 H2O', 'Glucose',   0.0200, chemicals),
    Rxn('Glucose + 2 H2O -> 2 Glycerol + O2',                       'Glucose',   0.0040, chemicals),
    Rxn('Glucose + 2 CO2 -> 2 SuccinicAcid + O2',                   'Glucose',   0.0060, chemicals),
    Rxn('3 Xylose -> 5 Ethanol + 5 CO2',                            'Xylose',    0.8500, chemicals),
    Rxn('Xylose + 0.039 CSL + 0.015 DAP -> 5 Z_mobilis + 2 H2O',
                                                                    'Xylose',    0.0190, chemicals),
    Rxn('3 Xylose + 5 H2O -> 5 Glycerol + 2.5 O2',                  'Xylose',    0.0030, chemicals),
    Rxn('Xylose + H2O -> Xylitol + 0.5 O2',                         'Xylose',    0.0460, chemicals),
    Rxn('3 Xylose + 5 CO2 -> 5 SuccinicAcid + 2.5 O2',              'Xylose',    0.0090, chemicals),
    ])
        self.CSL_to_constituents = Rxn(
            'CSL -> 0.5 H2O + 0.25 LacticAcid + 0.25 Protein', 'CSL', 1.0000, chemicals, basis='wt',
        )
        self.CSL_to_constituents.basis = 'mol'
        
    def _run(self):
        feed, *other = self.ins
        vent, effluent, sidedraw = self.outs
        effluent.copy_like(feed)
        vent.P = effluent.P = self.P
        vent.T = effluent.T = self.T_fermentation
        vent.phase = 'g'
        self.saccharification(effluent)
        self._batch_duty = effluent.Hnet - feed.Hnet
        sidedraw.copy_like(effluent)
        sidedraw.mol *= self.saccharification_split
        effluent.mol -= sidedraw.mol
        effluent.mix_from([effluent, *other], energy_balance=False)
        self.loss(effluent)
        self.cofermentation(effluent)
        self.CSL_to_constituents(effluent)
        vent.empty()
        vent.receive_vent(effluent, energy_balance=False)

    def _design(self):
        effluent = self.outs[1]
        v_0 = effluent.F_vol
        Design = self.design_results
        Design['Flow rate'] = v_0 / self.N_reactors
        tau = self.tau_saccharification + self.tau_cofermentation
        Design.update(size_batch(v_0, tau, self.tau_0, self.N_reactors, self.V_wf))
        Design['Batch duty'] = batch_duty = self._batch_duty
        Design['Reactor duty'] = reactor_duty = self.Hnet - batch_duty
        self.add_heat_utility(reactor_duty + batch_duty, effluent.T)
   


@cost('Flow rate', 'Recirculation pumps', kW=30, S=340*_gpm2m3hr,
      cost=47200, n=0.8, BM=2.3, CE=522, N='N_reactors')
@cost('Reactor duty', 'Heat exchangers', CE=522, cost=23900,
      S=-5*_Gcal2kJ, n=0.7, BM=2.2, N='N_reactors') # Based on a similar heat exchanger
@cost('Reactor volume', 'Agitators', CE=522, cost=52500,
      S=1e6*_gal2m3, n=0.5, kW=90, BM=1.5, N='N_reactors')
@cost('Reactor volume', 'Reactors', CE=522, cost=844000,
      S=1e6*_gal2m3, n=0.5, BM=1.5, N='N_reactors')
class SimultaneousSaccharificationAndCoFermentation(Unit):
    _N_ins = 1
    _N_outs = 2
    _ins_size_is_fixed = False
        
    #: Saccharification temperature (K)
    T_saccharification = 48+273.15
    
    #: Fermentation temperature (K)
    T_fermentation = 32+273.15
    
    #: Saccharification and Co-Fermentation time (hr)
    tau = 72
    
    #: Unload and clean up time (hr)
    tau_0 = 4
    
    #: Working volume fraction (filled tank to total tank volume)
    V_wf = 0.9
    
    #: Number of reactors
    N_reactors = 12
    
    _units = {'Flow rate': 'm3/hr',
              'Reactor volume': 'm3',
              'Batch duty': 'kJ/hr',
              'Reactor duty': 'kJ/hr'}
    
    def __init__(self, ID='', ins=None, outs=(), P=101325, thermo=None,
                 saccharification=None, loss=None, cofermentation=None):
        Unit.__init__(self, ID, ins, outs, thermo)
        self.P = P
        chemicals = self.chemicals
        #: [ParallelReaction] Enzymatic hydrolysis reactions including from 
        #: downstream batch tank in co-fermentation.
        self.saccharification = saccharification or ParallelRxn([
            Rxn('Glucan -> GlucoseOligomer',          'Glucan',   0.0400, chemicals),
            Rxn('Glucan + 0.5 H2O -> 0.5 Cellobiose', 'Glucan',   0.0120, chemicals),
            Rxn('Glucan + H2O -> Glucose',            'Glucan',   0.9000, chemicals),
            Rxn('Cellobiose + H2O -> 2Glucose',       'Cellobiose',  1.0000, chemicals)
        ])
        self.loss = loss or ParallelRxn([
        #   Reaction definition               Reactant    Conversion
        Rxn('Glucose -> 2 LacticAcid',       'Glucose',   0.0300, chemicals),
        Rxn('3 Xylose -> 5 LacticAcid',      'Xylose',    0.0300, chemicals),
        Rxn('3 Arabinose -> 5 LacticAcid',   'Arabinose', 0.0300, chemicals),
        Rxn('Galactose -> 2 LacticAcid',     'Galactose', 0.0300, chemicals),
        Rxn('Mannose -> 2 LacticAcid',       'Mannose',   0.0300, chemicals),])
        self.cofermentation = cofermentation or ParallelRxn([
        #   Reaction definition                                          Reactant    Conversion
        Rxn('Glucose -> 2 Ethanol + 2 CO2',                             'Glucose',   0.9500, chemicals),
        Rxn('Glucose + 0.047 CSL + 0.018 DAP -> 6 Z_mobilis + 2.4 H2O', 'Glucose',   0.0200, chemicals),
        Rxn('Glucose + 2 H2O -> 2 Glycerol + O2',                       'Glucose',   0.0040, chemicals),
        Rxn('Glucose + 2 CO2 -> 2 SuccinicAcid + O2',                   'Glucose',   0.0060, chemicals),
        Rxn('3 Xylose -> 5 Ethanol + 5 CO2',                            'Xylose',    0.8500, chemicals),
        Rxn('Xylose + 0.039 CSL + 0.015 DAP -> 5 Z_mobilis + 2 H2O',
                                                                        'Xylose',    0.0190, chemicals),
        Rxn('3 Xylose + 5 H2O -> 5 Glycerol + 2.5 O2',                  'Xylose',    0.0030, chemicals),
        Rxn('Xylose + H2O -> Xylitol + 0.5 O2',                         'Xylose',    0.0460, chemicals),
        Rxn('3 Xylose + 5 CO2 -> 5 SuccinicAcid + 2.5 O2',              'Xylose',    0.0090, chemicals),
        ])
        self.CSL_to_constituents = Rxn(
            'CSL -> 0.5 H2O + 0.25 LacticAcid + 0.25 Protein', 'CSL', 1.0000, basis='wt',
        )
        self.CSL_to_constituents.basis = 'mol'
        
    def _run(self):
        vent, effluent = self.outs
        vent.P = effluent.P = self.P
        vent.T = effluent.T = self.T_fermentation
        vent.phase = 'g'
        effluent.mix_from(self.ins, energy_balance=False)
        self.saccharification(effluent)
        self.loss(effluent)
        self.cofermentation(effluent)
        self.CSL_to_constituents(effluent)
        vent.empty()
        vent.receive_vent(effluent, energy_balance=False)

    def _design(self):
        effluent = self.outs[1]
        v_0 = effluent.F_vol
        Design = self.design_results
        Design['Flow rate'] = v_0 / self.N_reactors
        Design.update(size_batch(v_0, self.tau, self.tau_0, self.N_reactors, self.V_wf))
        Design['Reactor duty'] = reactor_duty = self.Hnet
        self.add_heat_utility(reactor_duty, effluent.T)


# %% Pretreatment separations 

# Membrane separation processes. Perry's Chemical Engineer's Handbook 7th Edition. 
@cost('Flow rate', 'Nanofiltration', kW=18000, S=0.25, units='m3/s',
      cost=17300, n=0.6, BM=1., CE=bst.units.design_tools.CEPCI_by_year[1996])
@cost('Annual flow rate', 'Membranes and maintenance', S=1., units='m3/yr',
      cost=0.13, n=1., BM=1., CE=bst.units.design_tools.CEPCI_by_year[1996],
      annual=True)
class ReverseOsmosis(bst.SolidsSeparator):
    pass

Europe_investment_site_factor = 1.2
euro_to_dollar = 1.04
installation_cost = 115000 # euro
volume_treated = 10 * 24 * 30 * 18 # m3
membrane_area = 27.5 # m2
membrane_cost = 95 # euro / m2
cleaning_cost = 50 # euro / m2
maintenance_cost = installation_cost * 0.02
yearly_operating_hours = 8000
operating_cost = membrane_area * (membrane_cost + cleaning_cost) + maintenance_cost * yearly_operating_hours / (18 * 30 * 24)
operating_cost_per_volume_treated = operating_cost / volume_treated
US_operating_cost = euro_to_dollar * operating_cost_per_volume_treated / Europe_investment_site_factor
US_installation_cost = euro_to_dollar * installation_cost / Europe_investment_site_factor
CEPCI2013 = bst.units.design_tools.CEPCI_by_year[2013]
electricity_cost = 2e3 # euro / yr
electricity_price = 0.04 # euro / kWh
electricity_demand = electricity_cost / yearly_operating_hours / electricity_price

@cost('Flow rate', 'Nanofiltration', kW=electricity_demand, S=10, units='m3/hr',
      cost=US_installation_cost, n=0.6, BM=1., CE=CEPCI2013)
@cost('Annual flow rate', 'Membranes and maintenance', S=1., units='m3/yr',
      cost=US_operating_cost, n=1., BM=1., CE=CEPCI2013,
      annual=True)
class Nanofilter(bst.Unit):
    _N_ins = 1
    _N_outs = 2
    volume_reduction = 0.80
    lignin_retention = 0.60
    NaOH_retention = 0.10
    sugar_retention = 0.95
    sugars_IDs = ('Glucose', 'Xylose', 'Arabinose', 'Galactose', 'Mannose',
              'GlucoseOligomer', 'GalactoseOligomer', 'MannoseOligomer',
              'XyloseOligomer', 'ArabinoseOligomer', 'HMF', 'Furfural', 'Xylobiose')
    lignin_IDs = ('SolubleLignin',)
    retentate_only_IDs = ('Lignin', 'Glucan', 'Xylan', 'Arabinan', 'Galactan',
                          'Mannan', 'Solids', 'Ash')
    
    def _run(self):
        feed, = self.ins
        permeate, retentate = self.outs
        permeate.empty()
        retentate.empty()
        lignin_retention = self.lignin_retention
        sugar_retention = self.sugar_retention
        NaOH_retention = self.NaOH_retention
        volume_reduction = self.volume_reduction
        water = feed.imass['Water']
        permeate.imass['Water'] = volume_reduction * water
        retentate.imass['Water'] = feed.imass['Water'] - permeate.imass['Water']
        retentate.copy_flow(feed, self.retentate_only_IDs)
        not_permeate_unique = (*self.retentate_only_IDs, *self.lignin_IDs, *self.sugars_IDs, 'NaOH', 'Water')
        not_permeate_unique = set([self.chemicals[i].ID for i in not_permeate_unique])
        permeate_unique = tuple([i.ID for i in self.chemicals if i.ID not in not_permeate_unique])
        permeate.copy_flow(feed, permeate_unique)
        F_mass = feed.F_mass
        lignin_comps = feed.imass[self.lignin_IDs]
        sugar_comps = feed.imass[self.sugars_IDs]
        NaOH = feed.imass['NaOH']
        lignin = lignin_comps.sum()
        sugar = sugar_comps.sum()
        lignin_def = lignin_comps / lignin
        sugar_def = sugar_comps / sugar
        NaOH = feed.imass['NaOH']
        feed_mass = np.array([lignin, sugar, NaOH])
        z_mass = feed_mass / F_mass
        K = np.array([lignin_retention, sugar_retention, NaOH_retention])
        phi_guess = 0.5
        Fa = permeate.F_mass
        Fb = retentate.F_mass
        assert abs((Fa + Fb + lignin + sugar + NaOH) - F_mass) < 1e-6
        permeate_mass = feed_mass / water * (1 - K) * permeate.imass['Water']
        
        # phi = tmo.separations.compute_phase_fraction(z_mass, K, phi_guess, Fa/F_mass, Fb/F_mass)
        # x_retentate = z_mass / (phi * K + (1. - phi))
        # retentate_mass = x_retentate * (1. - phi) * F_mass
        
        permeate.imass[self.lignin_IDs] = permeate_mass[0] * lignin_def
        permeate.imass[self.sugars_IDs] = permeate_mass[1] * sugar_def
        permeate.imass['NaOH'] = permeate_mass[2]
        retentate.mol = feed.mol - permeate.mol
        assert (permeate.mol >= 0.).all()
        retentate.T = permeate.T = feed.T


# %% Simple unit operations

@cost('Flow rate', 'Pump', units='kg/hr',
      S=402194, CE=522, cost=22500, n=0.8, kW=74.57, BM=2.3)
class HydrolysatePump(Unit): pass

@cost('Flow rate', 'Tank', S=1171, units='kg/hr',
      CE=522, cost=196000, n=0.7, BM=2)
class AmmoniaStorageTank(Unit): pass

@cost('Flow rate', 'Agitator', S=410846, units='kg/hr',
      CE=522, cost=None, n=0.5, BM=1.5, kW=14.914)
@cost('Flow rate', 'Tank', S=410369, units='kg/hr',
      CE=522, cost=None, n=0.7, BM=2)
class AmmoniaReacidificationTank(Unit): pass

@cost('Flow rate', 'Tank', S=425878, units='kg/hr',
      CE=522, cost=636000, n=0.7, BM=1.8)
@cost('Flow rate', 'Pump', S=425878, units='kg/hr',
      CE=522, cost=26800, n=0.8, BM=2.3, kW=93.2125)
@cost('Flow rate', 'Agitator', S=425878, units='kg/hr',
      CE=522, cost=68300, n=0.5, BM=1.5, kW=14.914)
class BeerTank(Unit): pass

@cost('Flow rate', 'Pump', S=292407, units='kg/hr',
      CE=551, cost=25365, n=0.8, BM=2.3, kW=93.2125)
class BlowdownDischargePump(Unit): pass

@cost('Flow rate', 'Tank', S=1393, units='kg/hr',
      CE=522, cost=70000, n=0.7, BM=2.6)
@cost('Flow rate', 'Pump', S=1393, units='kg/hr',
      CE=522, cost=3000, n=0.8, BM=3.1, kW=0.37285)
@cost('Flow rate', 'Agitator', S=1393, units='kg/hr',
      CE=522, cost=21200, n=0.5, BM=1.5, kW=7.457)
class CSLStorageTank(Unit): pass

@cost('Flow rate', 'Tank', S=163, units='kg/hr',
      CE=522, cost=102000, n=0.7, BM=1.8)
@cost('Flow rate', 'Pump', S=163, units='kg/hr',
      CE=522, cost=3000, n=0.8, BM=3.1, kW=0.37735)
@cost('Flow rate', 'Agitator', S=163, units='kg/hr',
      CE=522, cost=9800, n=0.5, BM=1.5, kW=4.10135)
@cost('Flow rate', 'Bag unloader', S=163, units='kg/hr',
      CE=522, cost=30000, n=0.6, BM=1.7)
class DAPStorageTank(Unit): pass

@cost('Flow rate', 'System', S=94697, units='kg/hr',
      CE=522, cost=13329690, n=0.6, BM=1.7, kW=783)
class FeedStockHandling(Unit): pass

@cost('Flow rate', 'Tank', S=252891, units='kg/hr',
      CE=522, cost=511000, n=0.7, BM=2)
@cost('Flow rate', 'Pump', S=252891, units='kg/hr',
      CE=522, cost=30000, n=0.8, BM=1.7, kW=55.9)
@cost('Flow rate', 'Agitator', S=252891, units='kg/hr',
      CE=522, cost=90000, n=0.5, BM=1.5, kW=170)
class PretreatmentFlash(bst.Flash): pass

@cost('Duty', 'Heat exchanger', S=8, units='Gcal/hr',
      CE=522, cost=85000, n=0.7, BM=2.2, magnitude=True)
class HydrolysateHeatExchanger(bst.HXutility): pass

@cost('Duty', 'Heat Exchanger', S=8, units='Gcal/hr',
      CE=551, cost=92000, n=0.7, BM=2.2, magnitude=True)
class PretreatmentWasteHeater(bst.HXutility): pass

@cost('Duty', 'Heat Exchanger', S=2, units='Gcal/hr',
      CE=522, cost=34000, n=0.7, BM=2.2, magnitude=True)
class WasteVaporCondenser(bst.HXutility): pass

@cost('Flow rate', 'Pump', S=402194, units='kg/hr',
      CE=522, cost=22500, n=0.8, BM=2.3, kW=74.57)
class HydrolyzatePump(Unit): pass

@cost('Flow rate', 'Separator', S=39000, units='kg/hr',
      CE=522, cost=35000000, n=0.7, BM=1.7)
class HydrolyzateSolidLiquidSeparator(Unit): pass

@cost('Flow rate', 'Tank', S=410369, units='kg/hr',
      CE=522, cost=236000, n=0.7, BM=2)
@cost('Flow rate', 'Agitator', S=410369, units='kg/hr',
      CE=522, cost=21900, n=0.5, BM=1.5, kW=7.457)
class AmmoniaAdditionTank(bst.MixTank): pass

@cost('Flow rate', 'Mixer', S=157478, units='kg/hr',
      CE=522, cost=5000, n=0.5, BM=1)
class AmmoniaMixer(bst.Mixer): pass

@cost('Flow rate', 'Mixer', S=380000, units='kg/hr',
      CE=522, cost=109000, n=0.5, BM=1.7, kW=74.57)
class EnzymeHydrolysateMixer(bst.Mixer): pass

@cost('Flow rate', 'Mixer', S=136260, units='kg/hr',
      CE=522, cost=6000, n=0.5, BM=1)
class SulfuricAcidMixer(bst.Mixer): pass

@cost('Flow rate', 'Tank', S=264116, units='kg/hr',
      CE=522, cost=203000, n=0.7, BM=2)
@cost('Flow rate', 'Pump', S=264116, units='kg/hr',
      CE=551, cost=17480, n=0.8, BM=1.7, kW=55.9)
@cost('Flow rate', 'Agitator', S=264116, units='kg/hr',
      CE=522, cost=90000, n=0.5, BM=1.5, kW=170)
class OligomerConversionTank(Unit): pass

@cost('Flow rate', 'Pump', S=402195, units='kg/hr',
      CE=522, cost=None, n=0.8, BM=2.3, kW=44.742)
class ReacidifiedHydrolyzatePump(Unit): pass

@cost('Flow rate', 'Pump', S=43149, units='kg/hr',
      CE=522, cost=8200, n=0.8, BM=2.3, kW=7.457)
@cost('Flow rate', 'Tank', S=40414, units='kg/hr',
      CE=522, cost=439000, n=0.7, BM=1.8)
@cost('Flow rate', 'Agitator', S=40414, units='kg/hr',
      CE=522, cost=31800, n=0.5, BM=1.5, kW=11.3205)
class SeedHoldTank(Unit): pass

@cost('Flow rate', 'Tank', S=1981, units='kg/hr',
      CE=522, cost=96000, n=0.7, BM=1.5)
@cost('Flow rate', 'Pump', S=1981, units='kg/hr',
      CE=522, cost=7493, n=0.8, BM=2.3, kW=0.5)
class SulfuricAcidStorageTank(Unit): pass

@cost('Flow rate', 'Tank', S=1981, units='kg/hr',
      CE=551, cost=6210, n=0.7, BM=2)
@cost('Flow rate', 'Pump', S=3720, units='kg/hr',
      CE=522, cost=8000, n=0.8, BM=2.3, kW=1)
class SulfuricAcidTank(Unit): pass



# -*- coding: utf-8 -*-
"""
.. autofunction:: biorefineries.cellulosic.systems.create_cofermentation_system
.. autofunction:: biorefineries.cellulosic.systems.create_integrated_bioprocess_saccharification_and_cofermentation_system
.. autofunction:: biorefineries.cellulosic.systems.create_saccharification_system
.. autofunction:: biorefineries.cellulosic.systems.create_simultaneous_saccharification_and_cofermentation_system
.. autofunction:: biorefineries.cellulosic.systems.create_cellulosic_fermentation_system

"""
from . import cofermentation
from . import integrated_bioprocess
from . import saccharification
from . import simultaneous_saccharification_cofermentation
from . import general_interface

__all__ = (
    *cofermentation.__all__,
    *integrated_bioprocess.__all__,
    *saccharification.__all__,
    *simultaneous_saccharification_cofermentation.__all__,
    *general_interface.__all__,
)

from .cofermentation import *
from .integrated_bioprocess import *
from .saccharification import *
from .simultaneous_saccharification_cofermentation import *
from .general_interface import *



# -*- coding: utf-8 -*-
"""

.. contents:: :local:

Oilcane
-------
.. automodule:: biorefineries.cane.systems.oilcane

Sugarcane
---------
.. automodule:: biorefineries.cane.systems.sugarcane

Biodiesel
---------
.. autofunction:: biorefineries.cane.systems.create_oilcane_to_biodiesel_1g
.. autofunction:: biorefineries.cane.systems.create_oilcane_to_biodiesel_combined_1_and_2g_post_fermentation_oil_separation
.. autofunction:: biorefineries.cane.systems.create_acTAG_separation_system
.. autofunction:: biorefineries.cane.systems.create_oilcane_to_biodiesel_and_actag_combined_1_and_2g_post_fermentation_oil_separation

Fermentation
------------
.. automodule:: biorefineries.cane.systems.fermentation.add_urea_MgSO4_nutrients
.. automodule:: biorefineries.cane.systems.fermentation.add_urea_nutrient
.. automodule:: biorefineries.cane.systems.fermentation.create_sucrose_fermentation_system

Juicing
-------
.. autofunction:: biorefineries.cane.systems.juicing.create_feedstock_handling_system
.. autofunction:: biorefineries.cane.systems.juicing.create_juicing_system_without_treatment
.. autofunction:: biorefineries.cane.systems.juicing.create_juicing_system_up_to_clarification
.. autofunction:: biorefineries.cane.systems.juicing.create_juicing_system

Bagasse
-------
.. autofunction:: biorefineries.cane.systems.bagasse.create_bagasse_pelleting_system
.. autofunction:: biorefineries.cane.systems.bagasse.create_bagasse_drying_system

Lipid extraction
----------------
.. autofunction:: biorefineries.cane.systems.lipid_extraction.create_lipid_exctraction_system
.. autofunction:: biorefineries.cane.systems.lipid_extraction.create_post_fermentation_oil_separation_system

Pretreatment
------------
.. autofunction:: biorefineries.cane.systems.pretreatment.create_cane_combined_1_and_2g_pretreatment

Sugar
-----
.. autofunction:: biorefineries.cane.systems.sugar.create_sugar_crystallization_system

"""
from . import bagasse
from . import fermentation
from . import juicing 
from . import sugar
from . import oilcane 
from . import sugarcane 
from . import lipidcane
from . import biodiesel
from . import cellulosic_biodiesel
from . import biodiesel_actag
from . import cellulosic_biodiesel_actag

__all__ = (
    *bagasse.__all__,
    *fermentation.__all__,
    *juicing.__all__,
    *sugar.__all__,
    *oilcane.__all__,
    *sugarcane.__all__, 
    *lipidcane.__all__,
    *biodiesel.__all__,
    *cellulosic_biodiesel.__all__,
    *biodiesel_actag.__all__,
    *cellulosic_biodiesel_actag.__all__,
)

from .bagasse import *
from .fermentation import *
from .juicing import *
from .sugar import *
from .oilcane import *
from .sugarcane import *
from .lipidcane import *
from .biodiesel import *
from .cellulosic_biodiesel import *
from .biodiesel_actag import *
from .cellulosic_biodiesel_actag import *


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 21 11:05:24 2017
@author: Yoel
"""
from biorefineries.cane import systems 

__all__ = systems.__all__ 

from biorefineries.cane.systems import *


# -*- coding: utf-8 -*-
"""
"""
import biorefineries.oilcane as oc
import biosteam as bst
import numpy as np
import itertools
from functools import cache
import pandas as pd
import os
from thermosteam.utils import style_axis
from warnings import filterwarnings
from thermosteam.utils import GG_colors, colors
from matplotlib import pyplot as plt

results_folder = os.path.join(os.path.dirname(__file__), 'results')
indicator_colors = (
    colors.neutral, GG_colors.purple, GG_colors.orange, GG_colors.green,
    GG_colors.red, GG_colors.blue, GG_colors.yellow,
)
indicator_sizes = (3, 3, 3,
                   2, 2, 2,
                   1, 1, 1,)
indicators = (
    'IRR', # %
    'GWP', # kg-CO2-eq / L
    'BY', # Biodiesel yield [MT / hc]
)
indicator_configurations = {
    'IRR-GWP-BY': [1, 1, 1],
    'GWP-BY': [0, 1, 1],
    'IRR-BY': [1, 0, 1],
    'IRR-GWP': [1, 1, 0],
    'IRR': [1, 0, 0],
    'GWP': [0, 1, 0],
    'BY': [0, 0, 1],
}
indicator_configurations = {i: np.array(j).reshape([3, 1]) for i, j in indicator_configurations.items()}
lines = ('WT', 'Energycane', '1566')
maximize_indicator_guide = {
    indicators[0]: True,
    indicators[1]: False,
    indicators[2]: True,
}

def create_scenarios(ranges, n=5):
    linspaces = [np.linspace(*i, n) for i in ranges]
    scenarios = itertools.product(*linspaces)
    return np.array(list(scenarios))

@cache
def create_model(configuration=None):
    if configuration is None:
        configuration = 'O6'
    if configuration == 'mockup':
        model = bst.Model(None)
    else:
        oc.load(configuration)
        model = bst.Model(oc.sys)
    parameter = model.parameter
    metric = model.metric
    
    @parameter(bounds=(50, 100), units='% sugarcane', baseline=60)
    def set_oilcane_biomass_yield(oilcane_biomass_yield):
        if oc.feedstock.imass['Lipid'] / (oc.feedstock.F_mass - oc.feedstock.imass['Water']) > 0.001:
            oc.dry_biomass_yield = oc.baseline_dry_biomass_yield * oilcane_biomass_yield * 0.01
    
    @parameter(bounds=(50, 90), units='%', baseline=60)
    def set_fermentation_yield(fermentation_yield):
        fermentation_yield /= 100
        for i in oc.fermentation_rxnsys.reactions:
            i.X[:2] = fermentation_yield
            i.X[2:] = 0.99 - fermentation_yield
    
    parameter(oc.set_cane_glucose_yield, name='Glucose yield')
    parameter(oc.set_cane_xylose_yield, name='Xylose yield')
    parameter(oc.set_saccharification_oil_recovery, name='Oil recovery')

    metric(oc.IRR)
    metric(oc.GWP_biodiesel_displacement)
    metric(oc.biodiesel_yield)
    return model

def set_cane_line(line):
    df = oc.get_composition_data()
    try:
        line = df.loc[line]
    except:
        line = df.loc[int(line)]
    oil = line['Stem Oil (dw)']
    moisture = line['Water (wt)']
    fiber = line['Fiber (dw)']
    biomass = line['Dry biomass yield (MT/hc)']
    oc.composition_specification.oil = oil['Mean']
    oc.composition_specification.fiber = fiber['Mean']
    oc.composition_specification.moisture = moisture['Mean']
    oc.composition_specification.load_composition()
    oc.dry_biomass_yield = biomass['Mean']
    GWP_sugarcane = oc.GWP_characterization_factors['sugarcane']
    oc.feedstock.set_CF(
        oc.GWP, GWP_sugarcane * (0.9 * oc.baseline_dry_biomass_yield / oc.dry_biomass_yield + 0.10)
    )

indicator_results = None   
def evaluate_winners():
    global indicator_results
    filterwarnings('ignore')
    model = create_model()
    scenarios = create_scenarios([i.bounds for i in model.parameters])
    model.load_samples(scenarios)
    if indicator_results is None:
        indices = [i.index for i in model.metrics]
        indicator_results = {i: [] for i in indicators}
        for line in lines:
            set_cane_line(line)
            model.evaluate()
            for indicator, index in zip(indicators, indices):
                indicator_results[indicator].append(model.table[index].copy())
    indicator_winners = np.vstack([
        (np.argmax if maximize_indicator_guide[i] else np.argmin)(
            np.vstack(indicator_results[i]), axis=0, keepdims=True,
        )
        for i in indicators
    ])
    
    winner_results = {i: {j: [] for j in indicator_configurations} for i in lines}
    for n, line in enumerate(lines):
        score_board = indicator_winners == n
        for indicator, configuration in indicator_configurations.items():
            winner_results[line][indicator] = scenarios[np.where((score_board == configuration).all(axis=0))]
        
    for line in lines:
        line_results = winner_results[line]
        xlfile = os.path.join(results_folder, f"{line}_winner.xlsx")
        with pd.ExcelWriter(xlfile) as writer:
            for indicator in indicator_configurations:
                pd.DataFrame(
                    line_results[indicator], columns=[i.name for i in model.parameters]
                ).to_excel(
                    writer, 
                    sheet_name=indicator
                )
        
def plot_winner_lines():
    model = create_model()
    fig, axes = plt.subplots(len(lines))
    parameters = model.parameters
    x = list(range(len(parameters)))
    
    def normalize_parameter_values(values):
        lbs = np.array([i.bounds[0] for i in parameters])
        ubs = np.array([i.bounds[1] for i in parameters])
        return (values - lbs) / (ubs - lbs)
    
    for ax, line in zip(axes, lines):
        plt.sca(ax)
        xlfile = os.path.join(results_folder, f"{line}_winner.xlsx")
        for indicator, color, size in zip(indicator_configurations, indicator_colors, indicator_sizes):
            df = pd.read_excel(xlfile, index_col=0, sheet_name=indicator)
            values = normalize_parameter_values(df.values)
            for y in values: 
                plt.plot(x, y, c=color.RGBn, lw=size)
        plt.ylim([0, 1])
        style_axis(
            ax, xticks=x, yticks=[0, 1],
            xticklabels=[f"{i.bounds[0]}\n{i.name}\n[{i.units}]" for i in parameters],
            yticklabels=['Min', 'Max'],
        )
        plt.sca(ax._cached_ytwin)
        plt.xticks(x, [f"{i.bounds[1]}" for i in parameters])
    plt.subplots_adjust(right=0.92, hspace=1, top=0.9, bottom=0.10)
            
                


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

__all__ = (
    'OilExtractionSpecification', 
    'MockExtractionSpecification',
)

class MockExtractionSpecification:
    
    def load_bagasse_oil_recovery(self, bagasse_recovery):
        pass
      
    def load_juicing_oil_recovery(self, recovery):
        pass
    
    def load_microbial_oil_recovery(self, recovery):
        pass
    
    def load_specifications(self, 
            recovery=None,
            bagasse_recovery=None,\
        ):
        pass

class OilExtractionSpecification:
    """
    Create a OilExtractionSpecification object for setting process 
    specifications related to oil extraction.
    
    Parameters
    ----------
    system : System
        System associated to feedstocks.
    crushing_mill : ChemicalIndexer
        Defines extraction recovery as a material split.
        
    Notes
    -----
    Use the `load_specifications` method to load specifications.
    
    """
    
    __slots__ = (
        'feedstocks',
        'crushing_mill',
        'pressure_filter',
        'cellmass_centrifuge',
        'juice',
        'screw_press',
        'crushing_mill_oil_recovery',
        'bagasse_oil_recovery',
        'microbial_oil_recovery',
        'juice_oil_recovery',
    )
    
    def __init__(self, system, crushing_mill, pressure_filter=None, cellmass_centrifuge=None, juice=None, 
                 screw_press=None, microbial_oil_recovery=None, juice_oil_recovery=None):
        self.crushing_mill = crushing_mill #: [Unit] Defines extraction recovery as a material split.
        self.pressure_filter = pressure_filter #: [Unit] Defines extraction recovery from bagasse as a material split.
        self.screw_press = screw_press
        self.cellmass_centrifuge = cellmass_centrifuge
        self.juice = juice
        if crushing_mill is None: 
            self.crushing_mill_oil_recovery = None
        else:
            self.crushing_mill_oil_recovery = crushing_mill.isplit['Oil'].mean()
        if juice is None:
            self.juice_oil_recovery = juice_oil_recovery
        else:
            self.juice_oil_recovery = 1. if juice_oil_recovery is None else juice_oil_recovery
        if pressure_filter is None:
            self.bagasse_oil_recovery = None
        else:
            self.bagasse_oil_recovery = 1. - pressure_filter.isplit['Oil'].mean()
            if pressure_filter in cellmass_centrifuge.get_downstream_units(facilities=False):
                if juice is None: raise ValueError('must pass both juice stream and pressure filter unit')
                self.microbial_oil_recovery = 0.7 if microbial_oil_recovery is None else microbial_oil_recovery
                # Microbial oil is removed further downstream
                @pressure_filter.add_specification(run=True)
                def adjust_oil_recovery():
                    cellmass = self.cellmass_centrifuge.outs[0]
                    juice = self.juice
                    pressure_filter = self.pressure_filter
                    total_oil = pressure_filter.ins[0].imass['Oil']
                    juice_oil_fraction = juice.imass['Oil'] / total_oil
                    cellmass_oil_fraction = cellmass.imass['Oil'] / total_oil
                    bagasse_oil_fraction = 1. - juice_oil_fraction - cellmass_oil_fraction
                    pressure_filter.isplit['Oil'] = (1.
                        - juice_oil_fraction * self.juice_oil_recovery
                        - cellmass_oil_fraction * self.microbial_oil_recovery
                        - bagasse_oil_fraction * self.bagasse_oil_recovery
                    )
            else:
                if screw_press:
                    # Microbial oil is removed separately from bagasse oil
                    self.microbial_oil_recovery = 1. - screw_press.isplit['Oil'].mean()
                    
                    @cellmass_centrifuge.add_specification(run=True)
                    def adjust_oil_recovery():
                        cellmass_centrifuge = self.cellmass_centrifuge
                        juice = self.juice
                        total_oil = cellmass_centrifuge.ins[0].imass['Oil']
                        juice_oil_fraction = juice.imass['Oil'] / total_oil
                        bagasse_oil_fraction = pressure_filter.outs[1].imass['Oil'] / total_oil
                        cellmass_centrifuge.aqueous_isplit['Oil'] = (1.
                            - juice_oil_fraction * self.juice_oil_recovery
                            - bagasse_oil_fraction * self.bagasse_oil_recovery
                        )
                else:
                    # No microbial oil
                    self.microbial_oil_recovery = microbial_oil_recovery
      
    def load_juicing_oil_recovery(self, recovery):
        if self.crushing_mill is None: return
        self.crushing_mill_oil_recovery = recovery
        self.crushing_mill.isplit['Oil'] = 1. - recovery
    
    def load_bagasse_oil_recovery(self, recovery):
        if self.pressure_filter is None: return
        self.bagasse_oil_recovery = recovery 
        self.pressure_filter.isplit['Oil'] = 1. - self.bagasse_oil_recovery
    
    def load_microbial_oil_recovery(self, recovery):
        if self.cellmass_centrifuge is None: return
        self.microbial_oil_recovery = recovery 
        if self.screw_press: self.screw_press.isplit['Oil'] = 1. - recovery
    
    def load_specifications(self, 
            crushing_mill_oil_recovery=None,
            bagasse_oil_recovery=None,
            microbial_oil_recovery=None,
        ):
        """
        Load oil extraction specifications.

        Parameters
        ----------
        recovery : float, optional
            Oil extraction recovery.
        bagasse_recovery : float, optional
            Oil extraction recovery from bagasse.
        oil_content : float, optional
            Oil content of feedstocks [dry wt. %].

        """
        if crushing_mill_oil_recovery is None: 
            crushing_mill_oil_recovery = self.crushing_mill_oil_recovery
        if bagasse_oil_recovery is None:
            bagasse_oil_recovery = self.bagasse_oil_recovery
        if microbial_oil_recovery is None:
            microbial_oil_recovery = self.microbial_oil_recovery
        self.load_juicing_oil_recovery(crushing_mill_oil_recovery)
        self.load_bagasse_oil_recovery(bagasse_oil_recovery)
        self.load_microbial_oil_recovery(microbial_oil_recovery)
    



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import thermosteam as tmo
import biosteam as bst
from biosteam import main_flowsheet as f
from biosteam import SystemFactory
from biorefineries.biodiesel import (
    create_lipid_pretreatment_system as create_oil_pretreatment_system,
    create_transesterification_and_biodiesel_separation_system,
)
from .fermentation import create_cane_to_combined_1_and_2g_fermentation
from .lipid_extraction import create_lipid_exctraction_system
from .biodiesel_actag import create_acTAG_separation_system
from .. import units
from .. import streams as s

__all__ = (
    'create_oilcane_to_biodiesel_and_actag_combined_1_and_2g_post_fermentation_oil_separation',
)

@SystemFactory(
    ID='oilcane_sys',
    ins=[s.oilcane],
    outs=[s.biodiesel, s.crude_glycerol, s.acTAG],
)
def create_oilcane_to_biodiesel_and_actag_combined_1_and_2g_post_fermentation_oil_separation(ins, outs):
    oilcane, = ins
    biodiesel, crude_glycerol, acTAG = outs
    cofermentation = tmo.PRxn([
        tmo.Rxn('Glucose -> 2.04 Water + 1.67 CO2 + 0.106 AcetylDiOlein', 'Glucose', 0.156),
        tmo.Rxn('Glucose -> 2.1 Water + 1.72 CO2 + 0.075 TriOlein', 'Glucose', 0.165),
        tmo.Rxn('Glucose -> Cells', 'Glucose', 0.10, basis='wt').copy(basis='mol'),
        tmo.Rxn('Xylose -> 2.04 Water + 1.67 CO2 + 0.106 AcetylDiOlein', 'Xylose', 0.156),
        tmo.Rxn('Xylose -> 2.1 Water + 1.72 CO2 + 0.075 TriOlein', 'Xylose', 0.165),
        tmo.Rxn('Xylose -> Cells', 'Xylose', 0.10, basis='wt').copy(basis='mol'),
    ])
    oilcane_to_fermentation_sys = create_cane_to_combined_1_and_2g_fermentation('oilcane_to_fermentation_sys',
        ins=oilcane, 
        product_group='Lipid',
        titer=2.5,
        productivity=0.33,
        cofermentation_reactions=cofermentation,
        seed_train_reactions=cofermentation,
        CoFermentation=units.CoFermentation,
        SeedTrain=units.SeedTrain,
        include_scrubber=False,
        fed_batch=False,
        mockup=True
    )
    beer, lignin, condensate, pretreatment_wastewater, fiber_fines, bagasse_to_boiler = oilcane_to_fermentation_sys.outs
    
    fermentor = f(units.CoFermentation)
    fermentor.selectivity = 0.75
    fermentor.product_yield = 0.321
    
    @fermentor.add_specification(run=True)
    def update_selectivity_and_product_yield():
        selectivity = fermentor.selectivity
        product_yield = fermentor.product_yield
        X = cofermentation.X
        X[0] = X[3] = product_yield * selectivity
        X[1] = X[4] = product_yield * (1. - selectivity)
    
    lipid_exctraction_sys, ledct = create_lipid_exctraction_system(
        ins=beer,
        mockup=True,
        area=400,
        udct=True,
    )
    oil, cellmass, wastewater = lipid_exctraction_sys.outs
    
    splitter = ledct['U404']
    splitter.lipid_recovery = 0.7
    @splitter.add_specification(run=True)
    def adjust_lipid_recovery():
        total_lipid = fermentor.outs[1].imass['Lipid']
        free_lipid = fermentor.ins[0].imass['Lipid']
        x_free = free_lipid / total_lipid
        splitter.isplit['Lipid'] = 1. - (
            splitter.lipid_recovery * (1 - x_free) + x_free
        )
    
    acTAG_separation_sys, acdct = create_acTAG_separation_system(
        'acTAG_separation_sys', oil, [acTAG, ''], mockup=True, area=500, udct=True
    )
    acTAG, TAG = acTAG_separation_sys.outs
    
    oil_pretreatment_sys, oil_pretreatment_dct = create_oil_pretreatment_system(
        ins=TAG,
        outs=['', 'polar_lipids', ''],
        mockup=True,
        area=800,
        udct=True
    )
    oil, polar_lipids, wastewater_small = oil_pretreatment_sys.outs
    # flash = oil_pretreatment_dct['F3']
    # @flash.add_specification
    # def remelt():
    #     flash.outs[1].phase
    
    create_transesterification_and_biodiesel_separation_system(
        ins=oil, 
        outs=[biodiesel, crude_glycerol, ''],
        mockup=True,
        area=800,
    )
    
    s = f.stream
    u = f.unit
    bst.create_all_facilities(
        feedstock=s.bagasse,
        recycle_process_water_streams=(condensate,),
        HXN_kwargs=dict(
            ID=1000,
            ignored=lambda: [u.D801.boiler, u.D802.boiler, u.H803, u.H802, 
                             u.H801, u.H804, u.H806, u.H809, oil_pretreatment_dct['F3'],
                             acdct['C1'], acdct['C2'], ledct['F201']],
            Qmin=1e3,
            acceptable_energy_balance_error=0.01,
        ),
        CHP_kwargs=dict(area=700),
        WWT_kwargs=dict(area=600),
        area=900,
    )